# Copy of the mothballed rdflib_sqlite module by gjhiggins, see
# https://github.com/RDFLib/rdflib-sqlite/
#
# Modified by Staffan Malmgren to fix a couple of bugs

try:
    from sqlite3 import dbapi2 as sqlite3  # python 2.5
except ImportError:
    import warnings
    warnings.warn(
        "sqlite3 not found, is it installed?")
    __test__ = False

import re
import os
import sys
from rdflib.graph import QuotedGraph
from rdflib.graph import RDF
from rdflib.term import Literal
from rdfextras.store.REGEXMatching import REGEXTerm
from rdfextras.store.REGEXMatching import PYTHON_REGEX
from rdfextras.store.AbstractSQLStore import AbstractSQLStore, Graph
from rdfextras.store.AbstractSQLStore import extractTriple, unionSELECT
from rdfextras.store.AbstractSQLStore import ASSERTED_NON_TYPE_PARTITION
from rdfextras.store.AbstractSQLStore import ASSERTED_TYPE_PARTITION
from rdfextras.store.AbstractSQLStore import ASSERTED_LITERAL_PARTITION
from rdfextras.store.AbstractSQLStore import QUOTED_PARTITION
from rdfextras.store.AbstractSQLStore import table_name_prefixes
from rdfextras.store.AbstractSQLStore import TRIPLE_SELECT_NO_ORDER

import logging
log = logging.getLogger(__name__)
import rdflib

Any = None

rdflib.plugin.register(
    "SQLite", rdflib.store.Store,
    "ferenda.thirdparty.SQLite", "SQLite")


# User-defined REGEXP operator
def regexp(expr, item):
    r = re.compile(expr)
    return r.match(item) is not None


class SQLite(AbstractSQLStore):
    """
    SQLite store formula-aware implementation.  It stores its triples in
    the following partitions:

    - Asserted non rdf:type statements
    - Asserted rdf:type statements (in a table which models Class membership)
      The motivation for this partition is primarily query speed and
      scalability as most graphs will always have more rdf:type statements
      than others
    - All Quoted statements

    In addition it persists namespace mappings in a separate table
    """

    # Works around a py3 problem in SQLGenerator where a single "NULL"
    # string is encoded into a bytestring, then rendered as a SQL
    # parameter "b'NULL'", by not encoding it to a bytestring in the
    # first place
    def buildLitDTypeClause(self, obj, tableName):
        if isinstance(obj,Literal):
            return obj.datatype is not None and (
                "%s.objDatatype=" % (tableName) + "%s", [obj.datatype]
                ) or None
        else:
            return None

    def buildLitLanguageClause(self,obj,tableName):
        if isinstance(obj,Literal):
            return obj.language is not None and (
                "%s.objLanguage=" % (tableName) + "%s",[obj.language]
                ) or None
        else:
            return None

    # Works around a problem with sqlite3 where Literals without
    # language or datatype had these fields stored as the string
    # 'NULL', resulting in RDFLib literals
    def buildLiteralTripleSQLCommand(self, subject, predicate, obj, context, storeId):
        """
        Builds an insert command for literal triples (statements where the
        object is a Literal)
        """
        from rdfextras.utils.termutils import statement2TermCombination
        triplePattern = int(statement2TermCombination(subject, predicate, obj, context))
        literal_table = "%s_literal_statements" % storeId
        command = "INSERT INTO %s " % literal_table +" (subject,predicate,object,context,termComb,objLanguage,objDatatype) VALUES (%s, %s, %s, %s, %s,%s,%s)"
        return command, [
            self.normalizeTerm(subject),
            self.normalizeTerm(predicate),
            self.normalizeTerm(obj),
            self.normalizeTerm(context.identifier),
            triplePattern,
            isinstance(obj, Literal) and obj.language or None,
            isinstance(obj, Literal) and obj.datatype or None]

    def executeSQL(self, cursor, qStr, params=None, paramList=False):
        # Fixed version of SQLGenerator.executeSQL that respects None
        # (converting to the unquoted string null)
        if not params:
            cursor.execute(str(qStr))
        elif paramList:
            raise Exception("Not supported!")
        else:
            # params = tuple([not isinstance(item,int) and "'%s'" % item or item
            #                   for item in params])
            quotedparams = []
            for item in params:
                if isinstance(item,int):
                    quotedparams.append(item)
                elif item is None:
                    quotedparams.append("null")
                else:
                    quotedparams.append("'%s'" % item)
            querystr = str(qStr).replace('"',"'")
            cursor.execute(querystr%tuple(quotedparams))
    


    context_aware = True
    formula_aware = True  # really?
    transaction_aware = True  # really?
    regex_matching = PYTHON_REGEX
    autocommit_default = False
    _Store__node_pickler = None

    def open(self, home, create=False):
        """
        Opens the store specified by the configuration string. If
        create is ``True`` a store will be created if it does not already
        exist. If create is ``False`` and a store does not already exist
        an exception is raised. An exception is also raised if a store
        exists, but there are insufficient permissions to open the
        store."""
        if create:
            db = sqlite3.connect(home)
            c = db.cursor()
            c.execute(
                CREATE_ASSERTED_STATEMENTS_TABLE % (self._internedId))
            c.execute(
                CREATE_ASSERTED_TYPE_STATEMENTS_TABLE % (self._internedId))
            c.execute(
                CREATE_QUOTED_STATEMENTS_TABLE % (self._internedId))
            c.execute(
                CREATE_NS_BINDS_TABLE % (self._internedId))
            c.execute(
                CREATE_LITERAL_STATEMENTS_TABLE % (self._internedId))
            for tblName, indices in [
                (
                    "%s_asserted_statements",
                    [
                        ("%s_A_termComb_index", ('termComb',)),
                        ("%s_A_s_index", ('subject',)),
                        ("%s_A_p_index", ('predicate',)),
                        ("%s_A_o_index", ('object',)),
                        ("%s_A_c_index", ('context',)),
                    ],
                ),
                (
                    "%s_type_statements",
                    [
                        ("%s_T_termComb_index", ('termComb',)),
                        ("%s_member_index", ('member',)),
                        ("%s_klass_index", ('klass',)),
                        ("%s_c_index", ('context',)),
                    ],
                ),
                (
                    "%s_literal_statements",
                    [
                        ("%s_L_termComb_index", ('termComb',)),
                        ("%s_L_s_index", ('subject',)),
                        ("%s_L_p_index", ('predicate',)),
                        ("%s_L_c_index", ('context',)),
                    ],
                ),
                (
                    "%s_quoted_statements",
                    [
                        ("%s_Q_termComb_index", ('termComb',)),
                        ("%s_Q_s_index", ('subject',)),
                        ("%s_Q_p_index", ('predicate',)),
                        ("%s_Q_o_index", ('object',)),
                        ("%s_Q_c_index", ('context',)),
                    ],
                ),
                (
                    "%s_namespace_binds",
                    [
                        ("%s_uri_index", ('uri',)),
                    ],
                )]:
                for indexName, columns in indices:
                    c.execute("CREATE INDEX %s on %s (%s)" % (
                        indexName % self._internedId, tblName % (
                                    self._internedId), ','.join(columns)))
            if home == ":memory:":
                db.commit()
            else:
                c.close()
                db.commit()
                db.close()

        if home == ":memory:":
            self._db = db
        else:
            self._db = sqlite3.connect(home)
        self._db.create_function("regexp", 2, regexp)

        if os.path.exists(home) or home == ":memory:":
            c = self._db.cursor()
            c.execute("SELECT * FROM sqlite_master WHERE type='table'")
            tbls = [rt[1] for rt in c.fetchall()]
            c.close()
            for tn in [tbl % (self._internedId)
                            for tbl in table_name_prefixes]:
                if tn not in tbls:
                    sys.stderr.write("table %s Doesn't exist\n" % (tn))
                    # The database exists, but one of the partitions
                    # doesn't exist
                    return 0
            # Everything is there (the database and the partitions)
            return 1
        # The database doesn't exist - nothing is there
        # return -1

    def destroy(self, home):
        """
        FIXME: Add documentation
        """
        if home == ":memory:":
            self._db = None
        else:
            db = sqlite3.connect(home)
            c = db.cursor()
            for tblsuffix in table_name_prefixes:
                try:
                    c.execute('DROP table %s' % tblsuffix % (self._internedId))
                except Exception as errmsg:
                    print(("unable to drop table: %s, %s" % (
                              tblsuffix % (self._internedId), errmsg)))
                    # pass
            # Note, this only removes the associated tables for the closed
            # world universe given by the identifier
            # print("Destroyed Close World Universe" + \
            #       " %s (in SQLite database %s)" % (
            #        self.identifier,home))
            db.commit()
            c.close()
            db.close()

    def EscapeQuotes(self, qstr):
        """
        Standard SQL escaping required
        """
        return qstr.replace("'", "''") if qstr is not None else ''

    # This is overridden to leave unicode terms as is
    # Instead of converting them to ascii (the default behavior)
    def normalizeTerm(self, term):
        if isinstance(term, (QuotedGraph, Graph)):
            return term.identifier
        elif isinstance(term, Literal):
            return self.EscapeQuotes(term)
        elif term is None or isinstance(term, (list, REGEXTerm)):
            return term
        else:
            return term

    # Where Clause  utility Functions
    # The predicate and object clause builders are modified in order to
    # optimize subjects and objects utility functions which can take
    # lists as their last argument (object,predicate - respectively)
    def buildSubjClause(self, subject, tableName):
        if isinstance(subject, REGEXTerm):
            return " REGEXP (%s," + " %s)" % (
                tableName and '%s.subject' % tableName or 'subject'), [subject]
        elif isinstance(subject, list):
            clauseStrings = []
            paramStrings = []
            for s in subject:
                if isinstance(s, REGEXTerm):
                    clauseStrings.append(
                        " REGEXP (%s," + " %s)" % (tableName \
                            and '%s.subject' % tableName \
                            or 'subject') + " %s")
                    paramStrings.append(self.normalizeTerm(s))
                elif isinstance(s, (QuotedGraph, Graph)):
                    clauseStrings.append(
                        "%s=" % (tableName \
                            and '%s.subject' % tableName \
                            or 'subject') + "%s")
                    paramStrings.append(self.normalizeTerm(s.identifier))
                else:
                    clauseStrings.append(
                        "%s=" % (tableName \
                            and '%s.subject' % tableName \
                            or 'subject') + "%s")
                    paramStrings.append(self.normalizeTerm(s))
            return '(' + ' or '.join(clauseStrings) + ')', paramStrings
        elif isinstance(subject, (QuotedGraph, Graph)):
            return "%s=" % (
                tableName \
                and '%s.subject' % tableName \
                or 'subject') + "%s", [self.normalizeTerm(subject.identifier)]
        else:
            return subject is not None and "%s=" % (
                tableName and \
                '%s.subject' % tableName \
                or 'subject') + "%s", [subject] \
                or None

    # Capable of taking a list of predicates as well (in which case
    # subclauses are joined with 'OR')
    def buildPredClause(self, predicate, tableName):
        if isinstance(predicate, REGEXTerm):
            return " REGEXP (%s," + " %s)" % (
                tableName and \
                '%s.predicate' % tableName \
                or 'predicate'), [predicate]
        elif isinstance(predicate, list):
            clauseStrings = []
            paramStrings = []
            for p in predicate:
                if isinstance(p, REGEXTerm):
                    clauseStrings.append(" REGEXP (%s," + " %s)" % (
                        tableName \
                        and '%s.predicate' % tableName \
                        or 'predicate'))
                else:
                    clauseStrings.append("%s=" % (
                        tableName \
                        and '%s.predicate' % tableName \
                        or 'predicate') + "%s")
                paramStrings.append(self.normalizeTerm(p))
            return '(' + ' or '.join(clauseStrings) + ')', paramStrings
        else:
            return predicate is not None and "%s=" % (
                tableName \
                and '%s.predicate' % tableName \
                or 'predicate') + "%s", [predicate] or None

    # Capable of taking a list of objects as well (in which case subclauses
    # are joined with 'OR')
    def buildObjClause(self, obj, tableName):
        if isinstance(obj, REGEXTerm):
            return " REGEXP (%s," + " %s)" % (
                tableName \
                and '%s.object' % tableName \
                or 'object'), [obj]
        elif isinstance(obj, list):
            clauseStrings = []
            paramStrings = []
            for o in obj:
                if isinstance(o, REGEXTerm):
                    clauseStrings.append(" REGEXP (%s," + " %s)" % (
                        tableName \
                        and '%s.object' % tableName \
                        or 'object'))
                    paramStrings.append(self.normalizeTerm(o))
                elif isinstance(o, (QuotedGraph, Graph)):
                    clauseStrings.append("%s=" % (
                        tableName \
                        and '%s.object' % tableName \
                        or 'object') + "%s")
                    paramStrings.append(self.normalizeTerm(o.identifier))
                else:
                    clauseStrings.append("%s=" % (
                        tableName \
                        and '%s.object' % tableName \
                        or 'object') + "%s")
                    paramStrings.append(self.normalizeTerm(o))
            return '(' + ' or '.join(clauseStrings) + ')', paramStrings
        elif isinstance(obj, (QuotedGraph, Graph)):
            return "%s=" % (
                tableName \
                and '%s.object' % tableName \
                or 'object') + "%s", [self.normalizeTerm(obj.identifier)]
        else:
            return obj is not None and "%s=" % (
                tableName \
                and '%s.object' % tableName \
                or 'object') + "%s", [obj] or None

    def buildContextClause(self, context, tableName):
        context = context is not None \
                    and self.normalizeTerm(context.identifier) \
                    or context
        if isinstance(context, REGEXTerm):
            return " REGEXP (%s," + " %s)" % (
                    tableName \
                    and '%s.context' % tableName \
                    or 'context'), [context]
        else:
            return context is not None and "%s=" % (
                tableName \
                and '%s.context' % tableName \
                or 'context') + "%s", [context] or None

    def buildTypeMemberClause(self, subject, tableName):
        if isinstance(subject, REGEXTerm):
            return " REGEXP (%s," + " %s)" % (
                tableName \
                and '%s.member' % tableName \
                or 'member'), [subject]
        elif isinstance(subject, list):
            clauseStrings = []
            paramStrings = []
            for s in subject:
                clauseStrings.append("%s.member=" % tableName + "%s")
                if isinstance(s, (QuotedGraph, Graph)):
                    paramStrings.append(self.normalizeTerm(s.identifier))
                else:
                    paramStrings.append(self.normalizeTerm(s))
            return '(' + ' or '.join(clauseStrings) + ')', paramStrings
        else:
            return subject and "%s.member = " % (tableName) + "%s", [subject]

    def buildTypeClassClause(self, obj, tableName):
        if isinstance(obj, REGEXTerm):
            return " REGEXP (%s," + " %s)" % (
                tableName \
                and '%s.klass' % tableName \
                or 'klass'), [obj]
        elif isinstance(obj, list):
            clauseStrings = []
            paramStrings = []
            for o in obj:
                clauseStrings.append("%s.klass=" % tableName + "%s")
                if isinstance(o, (QuotedGraph, Graph)):
                    paramStrings.append(self.normalizeTerm(o.identifier))
                else:
                    paramStrings.append(self.normalizeTerm(o))
            return '(' + ' or '.join(clauseStrings) + ')', paramStrings
        else:
            return obj is not None \
                and "%s.klass = " % tableName + "%s", [obj] or None

    def triples(self, pattern, context=None):
        """
        A generator over all the triples matching pattern. Pattern can
        be any objects for comparing against nodes in the store, for
        example, RegExLiteral, Date? DateRange?

        quoted table:                <id>_quoted_statements

        asserted rdf:type table:     <id>_type_statements

        asserted non rdf:type table: <id>_asserted_statements

        triple columns: subject, predicate, object, context, termComb,
                        objLanguage, objDatatype

        class membership columns: member, klass, context, termComb

        FIXME:  These union all selects *may* be further optimized by joins

        """
        (subject, predicate, obj) = pattern
        quoted_table = "%s_quoted_statements" % self._internedId
        asserted_table = "%s_asserted_statements" % self._internedId
        asserted_type_table = "%s_type_statements" % self._internedId
        literal_table = "%s_literal_statements" % self._internedId
        c = self._db.cursor()

        parameters = []

        if predicate == RDF.type:
            # Select from asserted rdf:type partition and quoted table
            # (if a context is specified)
            clauseString, params = self.buildClause(
                'typeTable', subject, RDF.type, obj, context, True)
            parameters.extend(params)
            selects = [
                (
                  asserted_type_table,
                  'typeTable',
                  clauseString,
                  ASSERTED_TYPE_PARTITION
                ),
            ]

        elif isinstance(predicate, REGEXTerm) \
            and predicate.compiledExpr.match(RDF.type) \
            or not predicate:
            # Select from quoted partition (if context is specified),
            # literal partition if (obj is Literal or None) and asserted
            # non rdf:type partition (if obj is URIRef or None)
            selects = []
            if not self.STRONGLY_TYPED_TERMS \
                or isinstance(obj, Literal) \
                or not obj \
                or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm)):
                clauseString, params = self.buildClause(
                        'literal', subject, predicate, obj, context)
                parameters.extend(params)
                selects.append((
                  literal_table,
                  'literal',
                  clauseString,
                  ASSERTED_LITERAL_PARTITION
                ))
            if not isinstance(obj, Literal) \
                and not (isinstance(obj, REGEXTerm) \
                and self.STRONGLY_TYPED_TERMS) \
                or not obj:
                clauseString, params = self.buildClause(
                    'asserted', subject, predicate, obj, context)
                parameters.extend(params)
                selects.append((
                  asserted_table,
                  'asserted',
                  clauseString,
                  ASSERTED_NON_TYPE_PARTITION
                ))

            clauseString, params = self.buildClause(
                'typeTable', subject, RDF.type, obj, context, True)
            parameters.extend(params)
            selects.append(
                (
                  asserted_type_table,
                  'typeTable',
                  clauseString,
                  ASSERTED_TYPE_PARTITION
                )
            )

        elif predicate:
            # Select from asserted non rdf:type partition (optionally),
            # quoted partition (if context is speciied), and literal
            # partition (optionally)
            selects = []
            if not self.STRONGLY_TYPED_TERMS \
                or isinstance(obj, Literal) \
                or not obj \
                or (self.STRONGLY_TYPED_TERMS and isinstance(obj, REGEXTerm)):
                clauseString, params = self.buildClause(
                    'literal', subject, predicate, obj, context)
                parameters.extend(params)
                selects.append((
                  literal_table,
                  'literal',
                  clauseString,
                  ASSERTED_LITERAL_PARTITION
                ))
            if not isinstance(obj, Literal) \
                and not (isinstance(obj, REGEXTerm) \
                and self.STRONGLY_TYPED_TERMS) \
                or not obj:
                clauseString, params = self.buildClause(
                    'asserted', subject, predicate, obj, context)
                parameters.extend(params)
                selects.append((
                  asserted_table,
                  'asserted',
                  clauseString,
                  ASSERTED_NON_TYPE_PARTITION
                ))

        if context is not None:
            clauseString, params = self.buildClause(
                'quoted', subject, predicate, obj, context)
            parameters.extend(params)
            selects.append(
                (
                  quoted_table,
                  'quoted',
                  clauseString,
                  QUOTED_PARTITION
                )
            )

        q = self._normalizeSQLCmd(
            unionSELECT(selects, selectType=TRIPLE_SELECT_NO_ORDER))
        self.executeSQL(c, q, parameters)
        # NOTE: SQLite does not support ORDER BY terms that aren't integers,
        # so the entire result set must be iterated in order to be able to
        # return a generator of contexts
        tripleCoverage = {}
        result = c.fetchall()
        c.close()
        for rt in result:
            s, p, o, (graphKlass, idKlass, graphId) = \
                                    extractTriple(rt, self, context)
            contexts = tripleCoverage.get((s, p, o), [])
            contexts.append(graphKlass(self, idKlass(graphId)))
            tripleCoverage[(s, p, o)] = contexts

        for (s, p, o), contexts in list(tripleCoverage.items()):
            yield (s, p, o), (c for c in contexts)

CREATE_ASSERTED_STATEMENTS_TABLE = """
CREATE TABLE %s_asserted_statements (
    subject       text not NULL,
    predicate     text not NULL,
    object        text not NULL,
    context       text not NULL,
    termComb      tinyint unsigned not NULL)"""

CREATE_ASSERTED_TYPE_STATEMENTS_TABLE = """
CREATE TABLE %s_type_statements (
    member        text not NULL,
    klass         text not NULL,
    context       text not NULL,
    termComb      tinyint unsigned not NULL)"""

CREATE_LITERAL_STATEMENTS_TABLE = """
CREATE TABLE %s_literal_statements (
    subject       text not NULL,
    predicate     text not NULL,
    object        text,
    context       text not NULL,
    termComb      tinyint unsigned not NULL,
    objLanguage   varchar(3),
    objDatatype   text)"""

CREATE_QUOTED_STATEMENTS_TABLE = """
CREATE TABLE %s_quoted_statements (
    subject       text not NULL,
    predicate     text not NULL,
    object        text,
    context       text not NULL,
    termComb      tinyint unsigned not NULL,
    objLanguage   varchar(3),
    objDatatype   text)"""

CREATE_NS_BINDS_TABLE = """
CREATE TABLE %s_namespace_binds (
    prefix        varchar(20) UNIQUE not NULL,
    uri           text,
    PRIMARY KEY (prefix))"""



# flake8: noqa
from .elements import serialize
from .elements import deserialize
from .elements import AbstractElement
from .elements import UnicodeElement
from .elements import CompoundElement
from .elements import TemporalElement
from .elements import PredicateElement
from .elements import OrdinalElement
from .elements import Link
from .elements import LinkSubject
from .elements import Body
from .elements import Title
from .elements import Page
from .elements import Nav
from .elements import SectionalElement
from .elements import Section
from .elements import Subsection
from .elements import Subsubsection
from .elements import Paragraph
from .elements import Preformatted
from .elements import Heading
from .elements import Footnote
from .elements import OrderedList
from .elements import UnorderedList
from .elements import ListItem



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
"""Hanterar domslut (detaljer och referat) från Domstolsverket. Data
hämtas fran DV:s (ickepublika) FTP-server, eller fran lagen.nu."""

# system libraries (incl six-based renames)
from datetime import datetime
from ftplib import FTP
from time import mktime
import codecs
import itertools
import os
import re
import zipfile
from six import text_type as str
from six.moves.urllib_parse import urljoin
import tempfile

# 3rdparty libs
import pkg_resources
from rdflib import Namespace, URIRef, Graph, RDF
import requests
import lxml.html
from lxml import etree
from bs4 import BeautifulSoup, NavigableString

# my libs
from ferenda import Document, DocumentStore, Describer, WordReader, FSMParser
from ferenda.decorators import managedparsing, newstate
from ferenda import util
from ferenda.sources.legal.se.legalref import LegalRef, Link
from ferenda.elements import Body, Paragraph, CompoundElement, OrdinalElement, Heading

from ferenda.elements.html import Strong, Em
from . import SwedishLegalSource, SwedishCitationParser, RPUBL
DCT = Namespace(util.ns['dct'])
PROV = Namespace(util.ns['prov'])

# Objektmodellen för rättsfall:
#
# meta:
#  <http://localhost:8000/res/dv/nja/2009/s_695> a rpubl:Rattsfallsreferat;
#     owl:sameAs <http://localhost:8000/res/dv/nja/2009:68>,
#                <http://rinfo.lagrummet.se/publ/rf/nja/2009:68>;
# This should be owl:sameAs <http://rinfo.lagrummet.se/serie/rf/nja>
#     rpubl:rattsfallspublikation <http:///localhost:8000/coll/dv/nja>;
#     rpubl:arsutgava "2009";
#     rpubl:lopnummer "68";
#     rpubl:sidnummer "695";
#     rpubl:referatAvDomstolsavgorande <http://localhost:8000/res/dv/hd/t170-08/2009-11-04>;
#     rpubl:referatrubrik "Överföring av mönsterregistrering..."@sv;
#     dct:identifier "NJA 1987 s 187";
#     dct:bibliographicCitation "NJA 1987:68"
# This shld b owl:sameAs <http://rinfo.lagrummet.se/org/domstolsverket>
#     dct:publisher <http://localhost:8000/org/domstolsverket>;
#     dct:issued "2009-11-05"^^xsd:date.
#
#
#  <http://localhost:8000/res/dv/hd/t170-08/2009-11-04> a VagledandeDomstolsavgorande;
#     owl:sameAs <http://rinfo.lagrummet.se/publ/dom/hd/t_170-08/2009-11-04>;
#     rpubl:avgorandedatum "2009-11-04"^^xsd:date;
#     rpubl:domstolsavdelning "2";
#     rpubl:malnummer "T 170-08";
# rpubl:lagrum <http://localhost:8000/res/sfs/1970:485#P1>;
# rpubl:lagrum <http://localhost:8000/res/sfs/1970:485#P1a>;
# rpubl:lagrum <http://localhost:8000/res/sfs/1970:485#P2>;
# rpubl:lagrum <http://localhost:8000/res/sfs/1970:485#P5>;
# rpubl:lagrum <http://localhost:8000/res/sfs/1970:485#P31>;
# rpubl:lagrum <http://localhost:8000/res/sfs/1970:485#P32>;
#     dct:title "Överföring av mönsterregistrering..."@sv;
# shld be owl:sameAs <http://rinfo.lagrummet.se/org/hoegsta_domstolen>
#     dct:publisher <http://localhost:8000/org/hoegsta_domstolen>;
#     dct:issued "2009-11-05"^^xsd:date;
#     dct:subject <http://localhost:8000/concept/Mönsterrätt>;
#     dct:subject <http://localhost:8000/concept/Dubbelöverlåtelse>;
#     dct:subject <http://localhost:8000/concept/Formgivarrätt>;
#     dct:subject <http://localhost:8000/concept/Godtrosförvärv>;
#     dct:subject <http://localhost:8000/concept/Formgivning>;
# litteratur? dct:references t bnodes...
#
# uri: http://localhost:8000/res/dv/nja/2009/s_695 # hard to construct from "HDO/T170-08", requires a rdf lookup like .value(pred=RDF.type, object=RPUBL.Rattsfallsreferat)
# lang: sv
# body: [Paragraph(), Paragraph(), Paragraph(), ...]


class MaxDownloadsReached(Exception):
    pass


class DVStore(DocumentStore):

    """Customized DocumentStore.
    """

    def basefile_to_pathfrag(self, basefile):
        return basefile

    def pathfrag_to_basefile(self, pathfrag):
        return pathfrag

    def downloaded_path(self, basefile, version=None, attachment=None, suffix=None):
        if not suffix:
            if os.path.exists(self.path(basefile, "downloaded", ".doc")):
                suffix = ".doc"
            elif os.path.exists(self.path(basefile, "downloaded", ".docx")):
                suffix = ".docx"
            else:
                suffix = self.downloaded_suffix
        return self.path(basefile, "downloaded", suffix, version, attachment)

    def intermediate_path(self, basefile, version=None, attachment=None):
        return self.path(basefile, "intermediate", ".xml")

    def list_basefiles_for(self, action, basedir=None):
        if not basedir:
            basedir = self.datadir
        if action == "parse":
            # Note: This pulls everything into memory before first
            # value is yielded. A more nifty variant is at
            # http://code.activestate.com/recipes/491285/
            d = os.path.sep.join((basedir, "downloaded"))
            for x in sorted(itertools.chain(util.list_dirs(d, ".doc"),
                                            util.list_dirs(d, ".docx"))):
                suffix = os.path.splitext(x)[1]
                pathfrag = x[len(d) + 1:-len(suffix)]
                yield self.pathfrag_to_basefile(pathfrag)
        else:
            for x in super(DVStore, self).list_basefiles_for(action, basedir):
                yield x

# (ab)use the CitationClass, with it's useful parse_recursive method,
# to use a legalref based parser instead of a set of pyparsing
# grammars. 

class OrderedParagraph(Paragraph, OrdinalElement):
    def as_html(self, baseuri):
        element = super(Instans, self).as_xhtml(baseuri)
        element.set('id', self.ordinal)

class DomElement(CompoundElement):
    tagname = "div"
    prop = None
    def _get_classname(self):
        return self.__class__.__name__.lower()
    classname = property(_get_classname)

    def as_xhtml(self, baseuri):
        element = super(DomElement, self).as_xhtml(baseuri)
        if self.prop:
            # ie if self.prop = ('ordinal', 'dct:identifier'), then
            # dct:identifier = self.ordinal
            if hasattr(self, self.prop[0]) and getattr(self, self.prop[0]):
                element.set('content', getattr(self, self.prop[0]))
                element.set('property', self.prop[1])
        return element

class Delmal(DomElement):
    prop = ('ordinal', 'dct:identifier')
    
class Instans(DomElement):
    prop = ('court', 'dct:creator')
        
class Dom(DomElement):
    prop = ('malnr', 'dct:identifier')
    
class Domskal(DomElement): pass 
class Domslut(DomElement): pass # dct:author <- names of judges
class Betankande(DomElement): pass # dct:author <- referent
class Skiljaktig(DomElement): pass # dct:author <- name
class Tillagg(DomElement): pass # dct:author <- name
class Endmeta(DomElement): pass
        
class DV(SwedishLegalSource):
    alias = "dv"
    downloaded_suffix = ".zip"
    rdf_type = RPUBL.Rattsfallsreferat
    documentstore_class = DVStore
    namespaces = ('rdf',  # always needed
                  'xsi',  # XML Schema/RDFa validation
                  'dct',  # title, identifier, etc
                  'xsd',  # datatypes
                  'owl',  # : sameAs
                  'prov',
                  ('rpubl', 'http://rinfo.lagrummet.se/ns/2008/11/rinfo/publ#')
                  )
    # This is very similar to SwedishLegalSource.required_predicates,
    # only DCT.title has been changed to RPUBL.referatrubrik (and if
    # our validating function grokked that rpubl:referatrubrik
    # rdfs:isSubpropertyOf dct:title, we wouldn't need this). Also, we
    # removed dct:issued because there is no actual way of getting
    # this data (apart from like the file time stamps).  On further
    # thinking, we remove RPUBL.referatrubrik as it's not present (or
    # required) for rpubl:Rattsfallsnotis
    required_predicates = [RDF.type, DCT.identifier, PROV.wasGeneratedBy]
    
    DCT = Namespace(util.ns['dct'])
    sparql_annotations = "res/sparql/dv-annotations.rq"

    def get_default_options(self):
        opts = super(DV, self).get_default_options()
        opts['ftpuser'] = None
        opts['ftppassword'] = None
        opts['parsebodyrefs'] = True
        return opts

    def canonical_uri(self, basefile):
        # The canonical URI for HDO/B3811-03 should be
        # http://localhost:8000/res/dv/nja/2004s510. We can't know
        # this URI before we parse the document. Once we have, we can
        # find the first rdf:type = rpubl:Rattsfallsreferat (or
        # rpubl:Rattsfallsnotis) and get its url.
        #
        # FIXME: It would be simpler and faster to read
        # DocumentEntry(self.store.entry_path(basefile))['id'], but
        # parse does not yet update the DocumentEntry once it has the
        # canonical uri/id for the document.
        p = self.store.distilled_path(basefile)
        if not os.path.exists(p):
            raise ValueError("No distilled file for basefile %s at %s" % (basefile, p))

        with self.store.open_distilled(basefile) as fp:
            g = Graph().parse(data=fp.read())
        for uri, rdftype in g.subject_objects(predicate=self.ns["rdf"].type):
            if rdftype in (self.ns['rpubl'].Rattsfallsreferat, self.ns['rpubl'].Rattsfallsnotis):
                return str(uri)
        raise ValueError("Can't find canonical URI for basefile %s in %s" % (basefile, p))

    # we override make_document to avoid having it calling
    # canonical_uri prematurely
    def make_document(self, basefile=None):
        doc = Document()
        doc.basefile = basefile
        doc.meta = self.make_graph()
        doc.lang = self.lang 
        doc.body = Body()
        doc.uri = None # can't know this yet
        return doc

    # FIXME: store.list_basefiles_for("parse") must be fixed to handle two
    # different suffixes. Maybe store.downloaded_path() as well, so that
    # it returns .docx if a .docx file indeed exists, and .doc otherwise.
    # But this case (where documents can be in two (or more) formats depending
    # on age isn't uncommon, maybe DocumentStore should support it natively
    # (like with optional suffix parameter to download_path)?

    def download(self):
        # recurse =~ download everything, which we do if force is
        # specified OR if we've never downloaded before
        recurse = False

        if self.config.force or not self.config.lastdownload:
            recurse = True

        self.downloadcount = 0  # number of files extracted from zip files
                               # (not number of zip files)
        try:
            if self.config.ftpuser:
                self.download_ftp("", recurse,
                                  self.config.ftpuser,
                                  self.config.ftppassword)
            else:
                self.log.warning("Config variable ftpuser not set, downloading from secondary source (https://lagen.nu/dv/downloaded/) instead")
                self.download_www("", recurse)
        except MaxDownloadsReached:  # ok we're done!
            pass

    def download_ftp(self, dirname, recurse, user, password, connection=None):
        self.log.debug('Listing contents of %s' % dirname)
        lines = []
        if not connection:
            connection = FTP('ftp.dom.se')
            connection.login(user, password)

        connection.cwd(dirname)
        connection.retrlines('LIST', lines.append)

        for line in lines:
            parts = line.split()
            filename = parts[-1].strip()
            if line.startswith('d') and recurse:
                self.download(filename, recurse)
            elif line.startswith('-'):
                basefile = os.path.splitext(filename)[0]
                if dirname:
                    basefile = dirname + "/" + basefile
                # localpath = self.store.downloaded_path(basefile)
                localpath = self.store.path(basefile, 'downloaded/zips', '.zip')
                if os.path.exists(localpath) and not self.config.force:
                    pass  # we already got this
                else:
                    util.ensure_dir(localpath)
                    self.log.debug('Fetching %s to %s' % (filename,
                                                          localpath))
                    connection.retrbinary('RETR %s' % filename,
                                          # FIXME: retrbinary calls .close()?
                                          open(localpath, 'wb').write)
                    self.process_zipfile(localpath)
        connection.cwd('/')

    def download_www(self, dirname, recurse):
        url = 'https://lagen.nu/dv/downloaded/%s' % dirname
        self.log.debug('Listing contents of %s' % url)
        resp = requests.get(url)
        iterlinks = lxml.html.document_fromstring(resp.text).iterlinks()
        for element, attribute, link, pos in iterlinks:
            if link.startswith("/"):
                continue
            elif link.endswith("/") and recurse:
                self.download_www(link, recurse)
            elif link.endswith(".zip"):
                basefile = os.path.splitext(link)[0]
                if dirname:
                    basefile = dirname + basefile

                # localpath = self.store.downloaded_path(basefile)
                localpath = self.store.path(basefile, 'downloaded/zips', '.zip')
                if os.path.exists(localpath) and not self.config.force:
                    pass  # we already got this
                else:
                    absolute_url = urljoin(url, link)
                    self.log.debug('Fetching %s to %s' % (link, localpath))
                    resp = requests.get(absolute_url)
                    with self.store._open(localpath, "wb") as fp:
                        fp.write(resp.content)
                    self.process_zipfile(localpath)

    # eg. HDO_T3467-96.doc or HDO_T3467-96_1.doc
    re_malnr = re.compile(r'([^_]*)_([^_\.]*)()_?(\d*)(\.docx?)')
    # eg. HDO_T3467-96_BYTUT_2010-03-17.doc or
    #     HDO_T3467-96_BYTUT_2010-03-17_1.doc or
    #     HDO_T254-89_1_BYTUT_2009-04-28.doc (which is sort of the
    #     same as the above but the "_1" goes in a different place)
    re_bytut_malnr = re.compile(
        r'([^_]*)_([^_\.]*)_?(\d*)_BYTUT_\d+-\d+-\d+_?(\d*)(\.docx?)')
    re_tabort_malnr = re.compile(
        r'([^_]*)_([^_\.]*)_?(\d*)_TABORT_\d+-\d+-\d+_?(\d*)(\.docx?)')

    # temporary helper
    def process_all_zipfiles(self):
        self.downloadcount = 0
        zippath = self.store.path('', 'downloaded/zips', '')
        for zipfilename in util.list_dirs(zippath, suffix=".zip"):
            self.log.info("%s: Processing..." % zipfilename)
            self.process_zipfile(zipfilename)

    def process_zipfile(self, zipfilename):
        """Extract a named zipfile into appropriate documents"""
        removed = replaced = created = untouched = 0
        if not hasattr(self, 'downloadcount'):
            self.downloadcount = 0
        try:
            zipf = zipfile.ZipFile(zipfilename, "r")
        except zipfile.BadZipfile as e:
            self.log.error("%s is not a valid zip file: %s" % (zipfilename,e))
            return 
        for bname in zipf.namelist():
            if not isinstance(bname, str):  # py2
                # Files in the zip file are encoded using codepage 437
                name = bname.decode('cp437')
            else:
                name = bname
            if "_notis_" in name:
                base, suffix = os.path.splitext(name)
                segments = base.split("_")
                coll, year = segments[0], segments[1]
                # Extract this doc as a temp file -- we won't be
                # creating an actual permanent file, but let
                # extract_notis extract individual parts of this file
                # to individual basefiles
                fp = tempfile.NamedTemporaryFile("wb", suffix=suffix, delete=False)
                filebytes = zipf.read(bname)
                fp.write(filebytes)
                fp.close()
                tempname = fp.name
                r = self.extract_notis(tempname, year, coll)
                created += r[0]
                untouched += r[1]
                os.unlink(tempname)
            else:
                name = os.path.split(name)[1]
                if 'BYTUT' in name:
                    m = self.re_bytut_malnr.match(name)
                elif 'TABORT' in name:
                    m = self.re_tabort_malnr.match(name)
                else:
                    m = self.re_malnr.match(name)
                if m:
                    (court, malnr, opt_referatnr, referatnr, suffix) = (
                        m.group(1), m.group(2), m.group(3), m.group(4), m.group(5))
                    assert ((suffix == ".doc") or (suffix == ".docx")
                            ), "Unknown suffix %s in %r" % (suffix, name)
                    if referatnr:
                        basefile = "%s/%s_%s" % (court, malnr, referatnr)
                    elif opt_referatnr:
                        basefile = "%s/%s_%s" % (court, malnr, opt_referatnr)
                    else:
                        basefile = "%s/%s" % (court, malnr)

                    outfile = self.store.path(basefile, 'downloaded', suffix)

                    if "TABORT" in name:
                        self.log.info("%s: Removing" % basefile)
                        if not os.path.exists(outfile):
                            self.log.warning("%s: %s doesn't exist" % (basefile,
                                                                       outfile))
                        else:
                            os.unlink(outfile)
                        removed += 1
                    elif "BYTUT" in name:
                        self.log.info("%s: Replacing with new" % basefile)
                        if not os.path.exists(outfile):
                            self.log.warning("%s: %s doesn't exist" %
                                             (basefile, outfile))
                        replaced += 1
                    else:
                        self.log.info("%s: Unpacking" % basefile)
                        if os.path.exists(outfile):
                            untouched += 1
                            continue
                        else:
                            created += 1
                    if not "TABORT" in name:
                        data = zipf.read(bname)
                        with self.store.open(basefile, "downloaded", suffix, "wb") as fp:
                            fp.write(data)

                        # Make the unzipped files have correct timestamp
                        zi = zipf.getinfo(bname)
                        dt = datetime(*zi.date_time)
                        ts = mktime(dt.timetuple())
                        os.utime(outfile, (ts, ts))

                        self.downloadcount += 1
                    if self.config.downloadmax and self.downloadcount >= self.config.downloadmax:
                        raise MaxDownloadsReached()
                else:
                    self.log.warning('Could not interpret filename %r i %s' %
                                     (name, os.path.relpath(zipfilename)))
        self.log.debug('Processed %s, created %s, replaced %s, removed %s, untouched %s files' %
                       (os.path.relpath(zipfilename), created, replaced, removed, untouched))


    def extract_notis(self, docfile, year, coll="HDO"):
        def find_month_in_previous(basefile):
            # The big word file with all notises might not
            # start with a month name -- try to find out
            # current month by examining the previous notis
            # (belonging to a previous word file)

            self.log.warning("No month specified in %s, attempting to look in previous file" % basefile)
            # HDO/2009_not_26 -> HDO/2009_not_25
            tmpfunc = lambda x: str(int(x.group(0)) - 1)
            prev_basefile = re.sub('\d+$', tmpfunc, basefile)
            prev_path = self.store.intermediate_path(prev_basefile)
            avd_p = None
            if os.path.exists(prev_path):
                soup = BeautifulSoup(util.readfile(prev_path))
                tmp = soup.find(["w:p", "para"])
                if re_avdstart.match(tmp.get_text().strip()):
                    avd_p = tmp
            if not avd_p:
                raise RuntimeError("Cannot find value for month in %s (looked in %s" % (basefile, prev_path))
            return avd_p

        # Given a word document containing a set of "notisfall" from
        # either HD or HFD (earlier RegR), spit out a intermediate XML
        # file for each notis.
        if coll == "HDO":
            re_notisstart = re.compile("(?P<day>Den \d+:[ae]. |)(?P<ordinal>\d+)\s*\.\s*\((?P<malnr>\w\s\d+-\d+)\)", flags=re.UNICODE)
            re_avdstart = re.compile("(Januari|Februari|Mars|April|Maj|Juni|Juli|Augusti|September|Oktober|November|December)$")
        else: # REG / HFD
            re_notisstart = re.compile("[\w\: ]*Lnr:(?P<court>\w+) ?(?P<year>\d+) ?not ?(?P<ordinal>\d+)", flags=re.UNICODE)
            re_avdstart = None
        created = untouched = 0
        intermediatefile = os.path.splitext(docfile)[0] + ".xml"
        r = WordReader()
        intermediatefile, filetype = r.read(docfile, intermediatefile)
        if filetype == "docx":
            self._simplify_ooxml(intermediatefile, pretty_print=False)
            soup = BeautifulSoup(util.readfile(intermediatefile))
            soup = self._merge_ooxml(soup)
            p_tag = "w:p"
            xmlns = ' xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml" xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"'
        else:
            soup = BeautifulSoup(util.readfile(intermediatefile))
            p_tag = "para"
            xmlns = ''
        iterator = soup.find_all(p_tag)
        basefile = None
        fp = None
        avd_p = None
        day = None
        for p in iterator:
            t = p.get_text().strip()
            if re_avdstart:
                # keep track of current month, store that in avd_p
                m = re_avdstart.match(t)
                if m:
                    avd_p = p
                    continue

            m = re_notisstart.match(t)
            if m:
                ordinal = m.group("ordinal")
                try:
                    if m.group("day"):
                        day = m.group("day")
                    else:
                        # inject current day in the first text node of
                        # p (which should inside of a <emphasis
                        # role="bold" or equivalent).
                        subnode = None
                        # FIXME: is this a deprecated method?
                        for c in p.recursiveChildGenerator():
                            if isinstance(c, NavigableString):
                                c.string.replace_with(day + str(c.string))
                                break
                except IndexError:
                    pass

                previous_basefile = basefile
                basefile = "%(coll)s/%(year)s_not_%(ordinal)s" % locals()
                self.log.info("%s: Extracting from %s file" % (basefile, filetype))
                created += 1
                downloaded_path = self.store.path(basefile, 'downloaded', '.'+filetype)
                with self.store._open(downloaded_path, "w"): 
                    pass # just create an empty placeholder file
                if fp:
                    fp.write("</body>\n")
                    fp.close()
                    if filetype == "docx":
                        self._simplify_ooxml(self.store.intermediate_path(previous_basefile))
                util.ensure_dir(self.store.intermediate_path(basefile))
                fp = open(self.store.intermediate_path(basefile), "w")
                fp.write('<body%s>' % xmlns)
                if filetype != "docx":
                    fp.write("\n")
                if coll == "HDO" and not avd_p:
                    avd_p = find_month_in_previous(basefile)
                if avd_p:
                    fp.write(repr(avd_p))
            if fp:
                fp.write(repr(p))
                if filetype != "docx":
                    fp.write("\n")
        if fp: # should always be the case
            fp.write("</body>\n")
            fp.close()
            if filetype == "docx":
                self._simplify_ooxml(self.store.intermediate_path(basefile))
        else:
            self.log.error("%s/%s: No notis were extracted (%s)" %
                           (coll,year,docfile))
        return created, untouched

    re_delimSplit = re.compile("[;,] ?").split
    labels = {'Rubrik': DCT.description,
              'Domstol': DCT['creator'],  # konvertera till auktoritetspost
              'Målnummer': RPUBL['malnummer'],
              'Domsnummer': RPUBL['domsnummer'],
              'Diarienummer': RPUBL['diarienummer'],
              'Avdelning': RPUBL['domstolsavdelning'],
              'Referat': DCT['identifier'],
              'Avgörandedatum': RPUBL['avgorandedatum'],  # konvertera till xsd:date
              }

    # Metadata som kan innehålla noll eller flera poster.
    # Litteratur/sökord har ingen motsvarighet i RPUBL-vokabulären
    multilabels = {'Lagrum': RPUBL['lagrum'],
                   'Rättsfall': RPUBL['rattsfallshanvisning'],
                   # dct:references vore bättre, men sådana ska inte ha literalvärden
                   'Litteratur': DCT['relation'],
                   'Sökord': DCT['subject']
                   }

    # Listan härledd från containers.n3/rattsfallsforteckningar.n3 i
    # rinfoprojektets källkod - en ambitiösare lösning vore att
    # läsa in de faktiska N3-filerna i en rdflib-graf.
    publikationsuri = {'NJA': 'http://rinfo.lagrummet.se/ref/rff/nja',
                       'RH': 'http://rinfo.lagrummet.se/ref/rff/rh',
                       'MÖD': 'http://rinfo.lagrummet.se/ref/rff/mod',
                       'RÅ': 'http://rinfo.lagrummet.se/ref/rff/ra',
                       'RK': 'http://rinfo.lagrummet.se/ref/rff/rk',
                       'MIG': 'http://rinfo.lagrummet.se/ref/rff/mig',
                       'AD': 'http://rinfo.lagrummet.se/ref/rff/ad',
                       'MD': 'http://rinfo.lagrummet.se/ref/rff/md',
                       'FÖD': 'http://rinfo.lagrummet.se/ref/rff/fod'}

    domstolsforkortningar = {'ADO': 'http://lagen.nu/org/2008/arbetsdomstolen',
                             'HDO': 'http://lagen.nu/org/2008/hogsta-domstolen',
                             'HGO': 'http://lagen.nu/org/2008/gota-hovratt',
                             'HNN': 'http://lagen.nu/org/2008/hovratten-for-nedre-norrland',
                             'HON': 'http://lagen.nu/org/2008/hovratten-for-ovre-norrland',
                             'HSB': 'http://lagen.nu/org/2008/hovratten-over-skane-och-blekinge',
                             'HSV': 'http://lagen.nu/org/2008/svea-hovratt',
                             'HVS': 'http://lagen.nu/org/2008/hovratten-for-vastra-sverige',
                             'MDO': 'http://lagen.nu/org/2008/marknadsdomstolen',
                             'MIG': 'http://lagen.nu/org/2008/migrationsoverdomstolen',
                             'MÖD': 'http://lagen.nu/org/2008/miljooverdomstolen',
                             'REG': 'http://lagen.nu/org/2008/regeringsratten',
                             'KST': 'http://lagen.nu/org/2008/kammarratten-i-stockholm'}

    # This is information you can get from RDL, but we hardcode it for
    # now.
    slugs = {'arbetsdomstolen': 'ad',
             'domstolsverket': 'dv',
             'göta hovrätt': 'hgo',
             'högsta domstolen': 'hd',
             'högsta förvaltningsdomstolen': 'hfd',
             'hovrätten för nedre norrland': 'hnn',
             'hovrätten för övre norrland': 'hon',
             'hovrätten för västra sverige': 'hvs',
             'hovrätten över skåne och blekinge': 'hsb',
             'justitiekanslern': 'jk',
             'kammarrätten i göteborg': 'kgg',
             'kammarrätten i jönköping': 'kjo',
             'kammarrätten i stockholm': 'kst',
             'kammarrätten i sundsvall': 'ksu',
             'marknadsdomstolen': 'md',
             'migrationsöverdomstolen': 'mig',
             'miljööverdomstolen': 'mod',
             'mark- och miljööverdomstolen': 'mod',
             'patentbesvärsrätten': 'pbr',
             'rättshjälpsnämnden': 'rhn',
             'regeringsrätten': 'regr',
             'statens ansvarsnämnd': 'san',
             'svea hovrätt': 'hsv'}

    @managedparsing
    def parse(self, doc):
        if not hasattr(self, 'lagrum_parser'):
            self.lagrum_parser = LegalRef(LegalRef.LAGRUM)
        if not hasattr(self, 'rattsfall_parser'):
            self.rattsfall_parser = LegalRef(LegalRef.RATTSFALL)
        docfile = self.store.downloaded_path(doc.basefile)

        intermediatefile = self.store.intermediate_path(doc.basefile)
        r = WordReader()
        intermediatefile, filetype = r.read(docfile, intermediatefile)

        if filetype == "docx":
            self._simplify_ooxml(intermediatefile)

        with codecs.open(intermediatefile, encoding="utf-8") as fp:
            patchedtext, patchdesc = self.patch_if_needed(doc.basefile,
                                                          fp.read())
        # The second step is to mangle the crappy XML produced by
        # antiword (docbook) or Word 2007 (OOXML) into a nice pair of
        # structures. rawhead is a simple dict that we'll later transform
        # into a rdflib Graph. rawbody is a list of plaintext strings, each
        # representing a paragraph.
        #
        # long-term FIXME: WordReader should expose a unified
        # interface for handling both kinds of word files so that we
        # wouldn't need both parse_ooxml() and
        # parse_antiword_docbook(). This might require some other tool
        # than antiword for old .doc files, as this throws away a LOT
        # of info.
        if "not" in doc.basefile:
            rawhead, rawbody = self.parse_not(patchedtext, doc.basefile, filetype)
        elif filetype == "docx":
            rawhead, rawbody = self.parse_ooxml(patchedtext, doc.basefile)
        else:
            rawhead, rawbody = self.parse_antiword_docbook(patchedtext, doc.basefile)
        sanitized_head = self.sanitize_metadata(rawhead, doc.basefile)
        doc.uri = self.polish_metadata(sanitized_head, doc)
        if patchdesc:
            doc.meta.add((URIRef(doc.uri),
                          self.ns['ferenda'].patchdescription,
                          patchdesc))
        doc.body = self.format_body(rawbody)
        return True


    def parse_not(self, text, basefile, filetype):
        basefile_regex = re.compile("(?P<type>\w+)/(?P<year>\d+)_not_(?P<ordinal>\d+)")
        referat_templ = {'REG': 'RÅ %(year)s not %(ordinal)s',
                         'HDO': 'NJA %(year)s not %(ordinal)s',
                         'HFD': 'HFD %(year)s not %(ordinal)s'}

        head = {}
        body = []

        m = basefile_regex.match(basefile).groupdict()
        coll = m['type']
        head["Referat"] = referat_templ[coll] % m

        soup = BeautifulSoup(text)
        if filetype == "docx":
            ptag = "w:p"
            soup = self._merge_ooxml(soup)
        else:
            ptag = "para"

        iterator = soup.find_all(ptag)
        if coll == "HDO":
            # keep in sync w extract_notis
            re_notisstart = re.compile("(?:Den (?P<avgdatum>\d+):[ae].\s+|)(?P<ordinal>\d+)\. ?\((?P<malnr>\w \d+-\d+)\)", flags=re.UNICODE)
            re_avgdatum = re_malnr = re_notisstart
            re_lagrum = re_sokord = None
            # headers consist of the first two chunks (month, then
            # date+ordinal+malnr)
            header = iterator.pop(0), iterator[0] # need to re-read
                                                  # the second chunk
                                                  # later
            curryear = m['year']
            currmonth = self.swedish_months[header[0].get_text().strip().lower()]
        else: # "REG", "HFD"
            # keep in sync like above
            re_notisstart = re.compile("[\w\: ]*Lnr:(?P<court>\w+) ?(?P<year>\d+) ?not ?(?P<ordinal>\d+)")
            re_malnr = re.compile(r"D:(?P<malnr>\d+\-\d+)")
            re_avgdatum = re.compile(r"[AD]:(?P<avgdatum>\d+\-\d+\-\d+)")
            re_sokord = re.compile("Uppslagsord: (?P<sokord>.*)", flags=re.DOTALL)
            re_lagrum = re.compile("Lagrum: ?(?P<lagrum>.*)", flags=re.DOTALL)
            # headers consists of the first five or six
            # chunks. Doesn't end until "^Not \d+."
            header = []
            done = False
            while not done:
                if re.match("Not \d+\. ", iterator[0].get_text().strip()):
                    done = True
                else:
                    tmp = iterator.pop(0)
                    if tmp.get_text().strip():
                        # REG specialcase 
                        if header and header[-1].get_text() == "Lagrum:":
                            header[-1].append(list(tmp.children)[0])
                        else:
                            header.append(tmp)
            
        if coll == "HDO":
            head['Domstol'] = "Högsta domstolen"
        elif coll == "HFD":
            head['Domstol'] = "Högsta förvaltningsdomstolen"
        elif coll == "REG":
            head['Domstol'] = "Regeringsrätten"
        else:
            raise ValueError("Unsupported: %s" % coll)
        for node in header:
            t = node.get_text()
            # if not malnr, avgdatum found, look for those
            for fld, key, rex in (('Målnummer', 'malnr', re_malnr),
                                  ('Avgörandedatum', 'avgdatum', re_avgdatum),
                                  ('Lagrum', 'lagrum', re_lagrum),
                                  ('Sökord', 'sokord', re_sokord)):
                if not rex: continue
                m = rex.search(t)
                if m and m.group(key):
                    if fld in ('Lagrum'): # Sökord is split by sanitize_metadata
                        head[fld] = self.re_delimSplit(m.group(key))
                    else:
                        head[fld] = m.group(key)

        if coll == "HDO" and 'Avgörandedatum' in head:
            head['Avgörandedatum'] = "%s-%02d-%02d" % (curryear, currmonth, int(head['Avgörandedatum']))

        # Do a basic conversion of the rest (bodytext) to Element objects
        #
        # This is generic enough that it could be part of WordReader
        for node in iterator:
            line = []
            if filetype == "doc":
                subiterator = node
            elif filetype == "docx":
                subiterator = node.find_all("w:r")
            for part in subiterator:
                if part.name:  
                    t = part.get_text()
                else:
                    t = str(part)  # convert NavigableString to pure string
                # if not t.strip():
                #     continue
                if filetype == "doc" and part.name == "emphasis": # docbook
                    if part.get("role") == "bold":
                        if line and isinstance(line[-1], Strong):
                            line[-1][-1] += t
                        else:
                            line.append(Strong([t]))
                    else:
                        if line and isinstance(line[-1], Em):
                            line[-1][-1] += t
                        else:
                            line.append(Em([t]))
                elif filetype == "docx" and part.find("w:rpr") and part.find("w:rpr").find(["w:b", "w:i"]): # ooxml
                    if part.rpr.b:
                        if line and isinstance(line[-1], Strong):
                            line[-1][-1] += t
                        else:
                            line.append(Strong([t]))
                    elif part.rpr.i:
                        if line and isinstance(line[-1], Em):
                            line[-1][-1] += t
                        else:
                            line.append(Em([t]))
                else:
                    if line and isinstance(line[-1], str):
                        line[-1] += t
                    else:
                        line.append(t)
            if line:
                body.append(line)
        return head, body
        
    def parse_ooxml(self, text, basefile):
        soup = BeautifulSoup(text)
        soup = self._merge_ooxml(soup)

        head = {}
        
        # Högst uppe på varje domslut står domstolsnamnet ("Högsta
        # domstolen") följt av referatnumret ("NJA 1987
        # s. 113").
        firstfield = soup.find("w:t")
        # Ibland ärdomstolsnamnet uppsplittat på två
        # w:r-element. Bäst att gå på all text i
        # föräldra-w:tc-cellen
        firstfield = firstfield.find_parent("w:tc")
        head['Domstol'] = firstfield.get_text(strip=True)

        nextfield = firstfield.find_next("w:tc")
        head['Referat'] = nextfield.get_text(strip=True)
        # Hitta övriga enkla metadatafält i sidhuvudet
        for key in self.labels:
            if key in head:
                continue
            node = soup.find(text=re.compile(key + ':'))
            if not node:
                # Sometimes these text fields are broken up
                # (eg "<w:t>Avgörand</w:t>...<w:t>a</w:t>...<w:t>tum</w:t>")
                # Use (ridiculous) fallback method
                if key not in ('Diarienummer', 'Domsnummer', 'Avdelning'): # not always present
                    self.log.warning("%s: Couldn't find field %r" % (basefile, key))
                continue

            txt = node.find_next("w:t").find_parent("w:p").get_text(strip=True)
            if txt:  # skippa fält med tomma strängen-värden
                head[key] = txt

        # Hitta sammansatta metadata i sidhuvudet
        for key in ["Lagrum", "Rättsfall"]:
            node = soup.find(text=re.compile(key + ':'))
            if node:
                textnodes = node.find_parent('w:tc').find_next_sibling('w:tc')
                if not textnodes:
                    continue
                items = []
                for textnode in textnodes.find_all('w:t'):
                    t = textnode.get_text(strip=True)
                    if t:
                        items.append(t)
                if items:
                    head[key] = items

        # The main text body of the verdict
        body = []
        for p in soup.find(text=re.compile('EFERAT')).find_parent('w:tr').find_next_sibling('w:tr').find_all('w:p'):
            ptext = ''
            for e in p.find_all("w:t"):
                ptext += e.string
            body.append(ptext)

        # Finally, some more metadata in the footer
        if soup.find(text=re.compile(r'Sökord:')):
            head['Sökord'] = soup.find(
                text=re.compile(r'Sökord:')).find_next('w:t').get_text(strip=True)

        if soup.find(text=re.compile('^\s*Litteratur:\s*$')):
            n = soup.find(text=re.compile('^\s*Litteratur:\s*$'))
            head['Litteratur'] = n.findNext('w:t').get_text(strip=True)
        return head, body

    def parse_antiword_docbook(self, text, basefile):
        soup = BeautifulSoup(text)
        head = {}
        header_elements = soup.find("para")
        header_text = ''
        for el in header_elements.contents:
            if hasattr(el, 'name') and el.name == "informaltable":
                break
            else:
                header_text += el.string

        # Högst uppe på varje domslut står domstolsnamnet ("Högsta
        # domstolen") följt av referatnumret ("NJA 1987
        # s. 113"). Beroende på worddokumentet ser dock XML-strukturen
        # olika ut. Det vanliga är att informationen finns i en
        # pipeseparerad paragraf:

        parts = [x.strip() for x in header_text.split("|")]
        if len(parts) > 1:
            head['Domstol'] = parts[0]
            head['Referat'] = parts[1]
        else:
            # alternativ står de på första raden i en informaltable
            row = soup.find("informaltable").tgroup.tbody.row.find_all('entry')
            head['Domstol'] = row[0].get_text(strip=True)
            head['Referat'] = row[1].get_text(strip=True)

        # Hitta övriga enkla metadatafält i sidhuvudet
        for key in self.labels:
            node = soup.find(text=re.compile(key + ':'))
            if node:
                txt = node.find_parent('entry').find_next_sibling('entry').get_text(strip=True)
                if txt:
                    head[key] = txt

        # Hitta sammansatta metadata i sidhuvudet
        for key in ["Lagrum", "Rättsfall"]:
            node = soup.find(text=re.compile(key + ':'))
            if node:
                head[key] = []
                textchunk = node.find_parent(
                    'entry').find_next_sibling('entry').string
                for line in [util.normalize_space(x) for x in textchunk.split("\n\n")]:
                    if line:
                        head[key].append(line)

        body = []
        for p in soup.find(text=re.compile('REFERAT')).find_parent('tgroup').find_next_sibling('tgroup').find('entry').get_text(strip=True).split("\n\n"):
            body.append(p)

        # Hitta sammansatta metadata i sidfoten
        head['Sökord'] = soup.find(text=re.compile('Sökord:')).find_parent(
            'entry').next_sibling.next_sibling.get_text(strip=True)

        if soup.find(text=re.compile('^\s*Litteratur:\s*$')):
            n = soup.find(text=re.compile('^\s*Litteratur:\s*$')).find_parent(
                'entry').next_sibling.next_sibling.get_text(strip=True)
            head['Litteratur'] = n
        return head, body

    # correct broken/missing metadata
    def sanitize_metadata(self, head, basefile):
        basefile_regex = re.compile('(?P<type>\w+)/(?P<year>\d+)-(?P<ordinal>\d+)')
        nja_regex = re.compile("NJA ?(\d+) ?s\.? ?(\d+) ?\( ?(?:NJA|) ?[ :]?(\d+) ?: ?(\d+)")
        date_regex = re.compile("(\d+)[^\d]+(\d+)[^\d]+(\d+)")
        referat_regex = re.compile("(?P<type>[A-ZÅÄÖ]+)[^\d]*(?P<year>\d+)[^\d]+(?P<ordinal>\d+)")
        referat_templ = {'ADO': 'AD %(year)s nr %(ordinal)s',
                         'AD': '%(type)s %(year)s nr %(ordinal)s',
                         'MDO': 'MD %(year)s:%(ordinal)s',
                         'NJA': '%(type)s %(year)s s. %(ordinal)s',
                         None: '%(type)s %(year)s:%(ordinal)s'
        }

        # 1. Attempt to fix missing Referat
        if not head.get("Referat"):
            # For some courts (MDO, ADO) it's possible to reconstruct a missing Referat from the basefile
            m = basefile_regex.match(basefile)
            if m and m.group("type") in ('ADO', 'MDO'):
                head["Referat"] = referat_templ[m.group("type")] % (m.groupdict())

        # 2. Correct known problems with Domstol not always being correctly specified
        if "Hovrättenför" in head["Domstol"] or "Hovrättenöver" in head["Domstol"]:
            head["Domstol"] = head["Domstol"].replace("Hovrätten", "Hovrätten ")
        try:
            # if this throws a KeyError, it's not canonically specified
            self.lookup_resource(head["Domstol"], cutoff=1)
        except KeyError:
            # lookup URI with fuzzy matching, then turn back to canonical label
            head["Domstol"] = self.lookup_label(str(self.lookup_resource(head["Domstol"])))

        # 3. Convert head['Målnummer'] to a list. Occasionally more than one
        # Malnummer is provided (c.f. AD 1994 nr 107, AD
        # 2005-117, AD 2003-111) in a comma, semicolo or space
        # separated list. AD 2006-105 even separates with " och ".
        #
        # HOWEVER, "Ö 2475-12" must not be converted to ["Ö2475-12"], not ['Ö', '2475-12']
        if head.get("Målnummer"):
            if head["Målnummer"][:2] in ('Ö ', 'B ', 'T '):
                head["Målnummer"] = [head["Målnummer"].replace(" ", "")]
            else:
                res = []
                for v in re.split("och|,|;|\s", head['Målnummer']):
                    if v.strip():
                        res.append(v.strip())
                head['Målnummer'] = res
        
        # 4. Create a general term for Målnummer or Domsnummer to act
        # as a local identifier
        if head.get("Målnummer"):
            head["_localid"] = head["Målnummer"]
        elif head.get("Domsnummer"):
            head["_localid"] = head["Domsnummer"]
        else:
            raise ValueError("Required key (Målnummer/Domsnummer) missing")

        # 5. For NJA, Canonicalize the identifier through a very
        # forgiving regex and split of the alternative identifier
        # as head['_nja_ordinal']
        #
        # "NJA 2008 s 567 (NJA 2008:86)"=>("NJA 2008 s 567", "NJA 2008:86")
        # "NJA 2011 s. 638(NJA2011:57)" => ("NJA 2011 s 638", "NJA 2001:57")
        # "NJA 2012 s. 16(2012:2)" => ("NJA 2012 s 16", "NJA 2012:2")
        if "NJA" in head["Referat"] and " not " not in head["Referat"]:
            m = nja_regex.match(head["Referat"])
            if m:
                head["Referat"] = "NJA %s s %s" % (m.group(1), m.group(2))
                head["_nja_ordinal"] = "NJA %s:%s" % (m.group(3), m.group(4))
            else:
                raise ValueError("Unparseable NJA ref '%s'" % head["Referat"])

        # 6 Canonicalize referats: Fix crap like "AD 2010nr 67",
        # "AD2011 nr 17", "HFD_2012 ref.58", "RH 2012_121", "RH2010
        # :180", "MD:2012:5", "MIG2011:14", "-MÖD 2010:32" and many
        # MANY more
        if " not " not in head["Referat"]: # notiser always have OK Referat
            m = referat_regex.search(head["Referat"])
            if m:
                if m.group("type") in referat_templ:
                    head["Referat"] = referat_templ[m.group("type")] % m.groupdict()
                else:
                    head["Referat"] = referat_templ[None] % m.groupdict()
            else:
                raise ValueError("Unparseable ref '%s'" % head["Referat"])

        # 7. Convert Sökord string to an actual list 
        res = []
        if head.get("Sökord"):
            for s in self.re_delimSplit(head["Sökord"]):
                s = util.normalize_space(s)
                if not s:
                    continue
                # terms longer than 72 chars are not legitimate
                # terms. more likely descriptions. If a term has a - in
                # it, it's probably a separator between a term and a
                # description
                while len(s) >= 72 and " - " in s:
                    h, s = s.split(" - ", 1)
                    res.append(h)
                if len(s) < 72:
                    res.append(s)
            head["Sökord"] = res

        # 8. Convert Avgörandedatum to a sensible value in the face of
        # irregularities like '2010-11 30', '2011 03-23' '2011-
        # 01-27', '2009.08.28' or '07-12-28'
        m = date_regex.match(head["Avgörandedatum"])
        if m:
            if len(m.group(1)) < 4:
                if int(m.group(1) <= 80): # '80-01-01' => '1980-01-01',
                    year = '19' + m.group(1) 
                else:                     # '79-01-01' => '2079-01-01',
                    year = '20' + m.group(1) 
            else:
                year = m.group(1)
            head["Avgörandedatum"] = "%s-%s-%s" % (year, m.group(2), m.group(3))
        else:
            raise ValueError("Unparseable date %s" % head["Avgörandedatum"])

        # 9. Done!
        return head

    # create nice RDF from the sanitized metadata
    def polish_metadata(self, head, doc):

        def ref_to_uri(ref):
            # FIXME: We'd like to retire legalref and replace it with
            # pyparsing grammars.
            nodes = self.rattsfall_parser.parse(ref)
            assert isinstance(nodes[0], Link), "Couldn't make URI from ref %s" % ref
            uri = nodes[0].uri
            return localize_uri(uri)

        def dom_to_uri(domstol, malnr, avg):
            baseuri = self.config.url
            slug = self.slugs[domstol.lower()]
            # FIXME: We should create multiple urls if we have multiple malnummers?
            first_malnr = malnr[0]
            return "%(baseuri)sres/dv/%(slug)s/%(first_malnr)s/%(avg)s" % locals()

        def localize_uri(uri):
            if "publ/rattsfall" in uri:
                return uri.replace("http://rinfo.lagrummet.se/publ/rattsfall",
                                   self.config.url + "res/dv")
            elif "publ/sfs/" in uri:
                return uri.replace("http://rinfo.lagrummet.se/publ/sfs",
                                   self.config.url + "res/sfs")

        def split_nja(value):
            return [x[:-1] for x in value.split("(")]

        def sokord_uri(value):
            return self.config.url + "concept/%s" % util.ucfirst(value).replace(' ', '_')

        # 1. mint uris and create the two Describers we'll use
        refuri = ref_to_uri(head["Referat"])

            
        refdesc = Describer(doc.meta, refuri)
        
        domuri = dom_to_uri(head["Domstol"],
                            head["_localid"],
                            head["Avgörandedatum"])
        domdesc = Describer(doc.meta, domuri)

        # 2. convert all strings in head to proper RDF
        #
        # 
        for label, value in head.items():
            if label == "Rubrik":
                value = util.normalize_space(value)
                refdesc.value(self.ns['rpubl'].referatrubrik, value, lang="sv")
                domdesc.value(self.ns['dct'].title, value, lang="sv")

            elif label == "Domstol":
                domdesc.rel(self.ns['dct'].publisher, self.lookup_resource(value))
            elif label == "Målnummer":
                for v in value:
                    # FIXME: In these cases (multiple målnummer, which
                    # primarily occurs with AD), we should really
                    # create separate domdesc objects (there are two
                    # verdicts, just summarized in one document)
                    domdesc.value(self.ns['rpubl'].malnummer, v)
            elif label == "Domsnummer":
                domdesc.value(self.ns['rpubl'].domsnummer, value)
            elif label == "Diarienummer":
                domdesc.value(self.ns['rpubl'].diarienummer, value)
            elif label == "Avdelning":
                domdesc.value(self.ns['rpubl'].avdelning, value)
            elif label == "Referat":
                for pred, regex in {'rattsfallspublikation': r'([^ ]+)',
                                    'arsutgava': r'(\d{4})',
                                    'lopnummer': r'\d{4}(?:\:| nr | not )(\d+)',
                                    'sidnummer': r's.? ?(\d+)'}.items():
                    m = re.search(regex, value)
                    if m:
                        if pred == 'rattsfallspublikation':
                            # "NJA" -> "http://localhost:8000/coll/dv/nja"
                            # "RÅ" -> "http://localhost:8000/coll/dv/rå" <-- FIXME, should be .../dv/ra
                            uri = self.config.url + "coll/dv/" + m.group(1).lower()
                            refdesc.rel(self.ns['rpubl'][pred], uri)
                        else:
                            refdesc.value(self.ns['rpubl'][pred], m.group(1))
                refdesc.value(self.ns['dct'].identifier, value)

            elif label == "_nja_ordinal":
                refdesc.value(self.ns['dct'].bibliographicCitation,
                              value)
                m = re.search(r'\d{4}(?:\:| nr | not )(\d+)', value)
                if m:
                    refdesc.value(self.ns['rpubl'].lopnummer, m.group(1))

                refdesc.rel(self.ns['owl'].sameAs,
                            self.config.url + "res/dv/nja/" + value.split(" ")[1])

            elif label == "Avgörandedatum":
                domdesc.value(self.ns['rpubl'].avgorandedatum, self.parse_iso_date(value))

            elif label == "Lagrum":
                for i in value:  # better be list not string
                    for node in self.lagrum_parser.parse(i):
                        if isinstance(node, Link):

                            domdesc.rel(self.ns['rpubl'].lagrum,
                                        localize_uri(node.uri))
            elif label == "Rättsfall":
                for i in value:
                    for node in self.rattsfall_parser.parse(i):
                        if isinstance(node, Link):
                            domdesc.rel(self.ns['rpubl'].rattsfall,
                                        localize_uri(node.uri))
            elif label == "Litteratur":
                for i in value.split(";"):
                    domdesc.value(self.ns['dct'].relation, util.normalize_space(i))
            elif label == "Sökord":
                for s in value:
                    domdesc.rel(self.ns['dct'].subject, sokord_uri(s))

        # 3. mint some owl:sameAs URIs
        refdesc.rel(self.ns['owl'].sameAs, self.sameas_uri(refuri))
        domdesc.rel(self.ns['owl'].sameAs, self.sameas_uri(domuri))

        # 4. Add some same-for-everyone properties
        refdesc.rel(self.ns['dct'].publisher, self.lookup_resource('Domstolsverket'))
        if 'not' in head['Referat']:
            refdesc.rdftype(self.ns['rpubl'].Rattsfallsnotis)
        else:
            refdesc.rdftype(self.ns['rpubl'].Rattsfallsreferat)
        domdesc.rdftype(self.ns['rpubl'].VagledandeDomstolsavgorande)
        refdesc.rel(self.ns['rpubl'].referatAvDomstolsavgorande, domuri)
        refdesc.value(self.ns['prov'].wasGeneratedBy, self.qualified_class_name())
# FIXME: implement this
#        # 5. Create a meaningful identifier for the verdict itself (ie. "Göta hovrätts dom den 42 september 2340 i mål B 1234-42")
#        domdesc.value(self.ns['dct'].identifier, dom_to_identifier(head["Domstol"],
#                                                                   head["_localid"],
#                                                                   head["Avgörandedatum"])
#
        
        # 6. assert that we have everything we need
        
        # 7. done!
        return refuri


    def format_body(self, paras):
        b = Body()
        # paras is typically a list of strings, but can be a list of
        # lists, where the innermost list consists of strings
        # interspersed with Element objects.
        for x in paras:
            if isinstance(x, str):
                x = [x]
            b.append(Paragraph(x))

        
        # find and link references -- this increases processing time
        # 5-10x, so do it only if requested. For some types (NJA
        # notiser) we could consider finding references anyway, as
        # these documents do not have headnotes with fixed fields for
        # references to statutory law and caselaw -- if we don't parse
        # body, we don't have nothing.
        if self.config.parsebodyrefs:
            if not hasattr(self, 'ref_parser'):
                self.ref_parser = LegalRef(LegalRef.RATTSFALL, LegalRef.LAGRUM, LegalRef.FORARBETEN)
            citparser = SwedishCitationParser(self.ref_parser)
            b = citparser.parse_recursive(b)

        # convert the unstructured list of Paragraphs to a
        # hierarchical tree of instances, domslut, domskäl, etc
        b = self.structure_body(b)
        return b

    def structure_body(self, paras):
        courtnames = ["Linköpings tingsrätt",
                      "Lunds tingsrätt",
                      "Umeå tingsrätt",
                      "Stockholms tingsrätt", 

                      "Göta hovrätt",
                      "Hovrätten över Skåne och Blekinge",
                      "Hovrätten för Övre Norrland",
                      "Svea hovrätt",

                      "Högsta domstolen"]

        rx = ('(?P<court>Försäkringskassan|Migrationsverket) beslutade (därefter|) den (?P<date>\d+ \w+ \d+) att',
              '(A överklagade beslutet till |)(?P<court>(Förvaltningsrätten|Länsrätten|Kammarrätten) i \w+(| län)(|, migrationsdomstolen|, Migrationsöverdomstolen)|Högsta förvaltningsdomstolen) \((?P<date>\d+-\d+\d+)')
        instans_matchers = [re.compile(x, re.UNICODE) for x in rx]
            
        
        def is_delmal(parser):
            chunk = parser.reader.peek()
            return str(chunk) in ("I", "II", "III")

        def is_instans(parser):
            chunk = parser.reader.peek()
            strchunk = str(chunk)
            if strchunk in courtnames:
                return True
            # elif re.search('Migrationsverket överklagade migrationsdomstolens beslut', strchunk):
            #     return True
            elif parser._state_stack == ['body']:
                # if we're at root level, *anything* starts a new instans
                return True
            else:
                for sentence in split_sentences(strchunk):
                    for r in (instans_matchers):
                        if r.match(sentence):
                            return True
            return False

        def is_heading(parser):
            chunk = parser.reader.peek()
            strchunk = str(chunk)
            # a heading is reasonably short and does not end with a
            # period (or other sentence ending typography)
            return len(strchunk) < 140 and not (strchunk.endswith(".") or
                                                strchunk.endswith(":") or
                                                strchunk.startswith("”"))
                                            

        def is_betankande(parser):
            strchunk = str(parser.reader.peek())
            return strchunk == "Målet avgjordes efter föredragning."
            
        def is_dom(parser):
            res = is_domskal(parser)
            return res

        def is_domskal(parser):
            strchunk = str(parser.reader.peek())
            if strchunk == "Skäl":
                return True
            if re.match("(Tingsrätten|Hovrätten|HD|Högsta förvaltningsdomstolen) \([^)]*\) (meddelade|anförde|fastställde|yttrade)", strchunk):
                return True

        def is_domslut(parser):
            strchunk = str(parser.reader.peek())
            return strchunk in ("Domslut", "Hovrättens avgörande", "HD:s avgörande", "Högsta förvaltningsdomstolens avgörande")
            
        def is_skiljaktig(parser):
            strchunk = str(parser.reader.peek())
            return re.match("(Justitie|Kammarrätts)råde[nt] ([^\.]*) var (skiljaktig|av skiljaktig mening)", strchunk)

        def is_tillagg(parser):
            strchunk = str(parser.reader.peek())
            return re.match("Justitieråde[nt] ([^\.]*) (tillade för egen del|gjorde för egen del ett tillägg)", strchunk)

        def is_endmeta(parser):
            strchunk = str(parser.reader.peek())
            return re.match("HD:s (beslut|dom) meddela(de|d|t): den", strchunk)

        def is_paragraph(parser):
            return True

        # FIXME: This and make_paragraph ought to be expressed as
        # generic functions in the ferenda.fsmparser module
        @newstate('body')
        def make_body(parser):
            return parser.make_children(Body())

        @newstate('delmal')
        def make_delmal(parser):
            d = Delmal(ordinal=str(parser.reader.next()), malnr=None)
            return parser.make_children(d)

        @newstate('instans')
        def make_instans(parser):
            chunk = parser.reader.next()
            strchunk = str(chunk)
            if strchunk in courtnames:
                i = Instans(court=strchunk)
            else:
                i = False
                for sentence in split_sentences(strchunk):
                    for r in (instans_matchers):
                        m = r.match(sentence)
                        if m:
                            if 'court' in m.groupdict():
                                i = Instans([chunk], court=m.groupdict()['court'])
                            break
                    if i:
                        break
                if not i:
                    i = Instans([chunk])
                    
            return parser.make_children(i)

        def split_sentences(text):
            text = util.normalize_space(text)
            text += " "
            return text.split(". ")
                

        def make_heading(parser):
            # a heading is by definition a single line
            return Heading(parser.reader.next())

        @newstate('betankande')
        def make_betankande(parser):
            b = Betankande()
            b.append(parser.reader.next())
            return parser.make_children(b)

        @newstate('dom')
        def make_dom(parser):
            d = Dom(avgorandedatum=None, malnr=None)
            return parser.make_children(d)

        @newstate('domskal')
        def make_domskal(parser):
            d = Domskal()
            return parser.make_children(d)

        @newstate('domslut')
        def make_domslut(parser):
            d = Domslut()
            return parser.make_children(d)

        @newstate('skiljaktig')
        def make_skiljaktig(parser):
            s = Skiljaktig()
            s.append(parser.reader.next())
            return parser.make_children(s)

        @newstate('tillagg')
        def make_tillagg(parser):
            t = Tillagg()
            t.append(parser.reader.next())
            return parser.make_children(t)

        @newstate('endmeta')
        def make_endmeta(parser):
            m = Endmeta()
            m.append(parser.reader.next())
            return parser.make_children(m)
            
        
            
        def make_paragraph(parser):
            chunk = parser.reader.next()
            strchunk = str(chunk)
            if ordered(strchunk):
                # FIXME: Cut the ordinal from chunk somehow
                if isinstance(chunk, Paragraph):
                    p = OrderedParagraph(list(chunk), ordinal=ordered(strchunk))
                else:
                    p = OrderedParagraph([chunk], ordinal=ordered(strchunk))
            else:
                if isinstance(chunk, Paragraph):
                    p = chunk
                else: 
                    p = Paragraph([chunk])
            return p

        def ordered(chunk):
            if re.match("(\d+).", chunk):
                return chunk.split(".", 1)

        def transition_domskal(symbol, statestack):
            if 'betankande' in statestack:
                # Ugly hack: mangle the statestack so that *next time*
                # we encounter a is_domskal, we pop the statestack,
                # but for now we push to it.
                statestack[statestack.index('betankande')] = "__done__"
                return make_domskal, "domskal"
            else:
                # here's where we pop the stack
                return False, None
                
        p = FSMParser()
        p.set_recognizers(is_delmal,
                          is_endmeta,
                          is_instans,
                          is_dom,
                          is_betankande,
                          is_domskal,
                          is_domslut,
                          is_skiljaktig,
                          is_tillagg,
                          is_heading,
                          is_paragraph)
        commonstates = ("body", "delmal", "instans", "domskal", "domslut", "betankande", "skiljaktig", "tillagg")
        
        p.set_transitions({
            ("body", is_delmal): (make_delmal, "delmal"),
            ("body", is_instans): (make_instans, "instans"),
            ("body", is_endmeta): (make_endmeta, "endmeta"),
            ("delmal", is_instans): (make_instans, "instans"),
            ("delmal", is_delmal): (False, None),
            ("delmal", is_endmeta): (False, None),
            ("instans", is_betankande): (make_betankande, "betankande"),
            ("instans", is_dom): (make_dom, "dom"),
            ("instans", is_domslut): (make_domslut, "domslut"),
            ("instans", is_instans): (False, None),
            ("instans", is_skiljaktig): (make_skiljaktig, "skiljaktig"),
            ("instans", is_tillagg): (make_tillagg, "tillagg"),
            ("instans", is_delmal): (False, None),
            ("instans", is_endmeta): (False, None),
            ("betankande", is_domskal): transition_domskal, # either (make_domskal, "domskal") or (False, None)
            ("betankande", is_domslut): (make_domslut, "domslut"),
            ("__done__", is_domskal): (False, None), 
            ("__done__", is_skiljaktig): (False, None), 
            ("__done__", is_tillagg): (False, None), 
            ("__done__", is_delmal): (False, None), 
            ("__done__", is_endmeta): (False, None), 
            ("__done__", is_domslut): (make_domslut, "domslut"),
            ("dom", is_domskal): (make_domskal, "domskal"),
            ("dom", is_domslut): (make_domslut, "domslut"),
            ("dom", is_instans): (False, None),
            ("dom", is_skiljaktig): (False, None), # Skiljaktig mening is not considered
                                                   # part of the dom, but rather an appendix
            ("dom", is_tillagg): (False, None), 
            ("dom", is_endmeta): (False, None),
            ("domskal", is_delmal): (False, None), 
            ("domskal", is_domslut): (False, None),
            ("domskal", is_instans): (False, None), 
            ("domslut", is_instans): (False, None),
            ("domslut", is_domskal): (False, None),
            ("domslut", is_skiljaktig): (False, None), 
            ("domslut", is_tillagg): (False, None),
            ("domslut", is_endmeta): (False, None),
            ("skiljaktig", is_domslut): (False, None),
            ("skiljaktig", is_instans): (False, None),
            ("skiljaktig", is_skiljaktig): (False, None),
            ("skiljaktig", is_tillagg): (False, None),
            ("skiljaktig", is_delmal): (False, None),
            ("tillagg", is_tillagg): (False, None),
            ("tillagg", is_delmal): (False, None),
            ("tillagg", is_endmeta): (False, None),
            ("endmeta", is_paragraph): (make_paragraph, None),
            (commonstates, is_heading): (make_heading, None),
            (commonstates, is_paragraph): (make_paragraph, None),
                       })
        p.initial_state = "body"
        p.initial_constructor = make_body
        p.debug = os.environ.get('FERENDA_FSMDEBUG', False)
        return p.parse(paras)
        
    # FIXME: port to relate_all_setup / _teardown
    def GenerateMapAll(self):
        mapfile = os.path.sep.join(
            [self.baseDir, 'dv', 'generated', 'uri.map'])
        util.robust_remove(mapfile + ".new")

        parsed_dir = os.path.sep.join([self.baseDir, 'dv', 'parsed'])
        self._do_for_all(parsed_dir, '.xht2', self.GenerateMap)
        util.robustRename(mapfile + ".new", mapfile)

    def GenerateMap(self, basefile):
        start = time()
        infile = os.path.relpath(self._xmlFileName(basefile))
        head = codecs.open(infile, encoding='utf-8').read(1024)
        m = self.re_xmlbase(head)
        if m:
            uri = "http://rinfo.lagrummet.se/publ/rattsfall/%s" % m.group(1)
            mapfile = self.store.path('generated', 'uri.map', '.new')
            util.ensure_dir(mapfile)
            f = codecs.open(mapfile, 'a', encoding='iso-8859-1')
            f.write("%s\t%s\n" % (m.group(1), basefile))
            f.close()
            self.log.info("%s ok" % basefile)
            return
        else:
            self.log.warning("could not find xml:base in %s" % infile)

    def _simplify_ooxml(self, filename, pretty_print=True):
        # simplify the horrendous mess that is OOXML through simplify-ooxml.xsl
        with open(filename) as fp:
            intree = etree.parse(fp)
        fp = pkg_resources.resource_stream('ferenda', "res/xsl/simplify-ooxml.xsl")
        transform = etree.XSLT(etree.parse(fp))
        fp.close()
        resulttree = transform(intree)
        with open(filename, "wb") as fp:
            fp.write(etree.tostring(resulttree, pretty_print=pretty_print, encoding="utf-8"))
        

    def _merge_ooxml(self, soup):
        # this is a similar step to _simplify_ooxml, but merges w:p
        # elements in a BeautifulSoup tree. This step probably should
        # be performed through XSL and be put in _simplify_ooxml as
        # well.
                # The soup now contains a simplified version of OOXML where
        # lot's of nonessential tags has been stripped. However, the
        # central w:p tag often contains unneccessarily splitted
        # subtags (eg "<w:t>Avgörand</w:t>...<w:t>a</w:t>...
        # <w:t>tum</w:t>"). Attempt to join these
        #
        # FIXME: This could be a part of simplify_ooxml instead.
        for p in soup.find_all("w:p"):
            current_r = None
            for r in p.find_all("w:r"):
                # find out if formatting instructions (bold, italic)
                # are identical
                if current_r and current_r.find("w:rpr") == r.find("w:rpr"):
                    # ok, merge
                    ts = list(current_r.find_all("w:t"))
                    assert len(ts) == 1, "w:r should not contain exactly one w:t"
                    ns = ts[0].string
                    ns.replace_with(str(ns) + r.find("w:t").string)
                    r.decompose()
                else:
                    current_r = r
        return soup
    
            
    # gonna need this for news_criteria()
    pubs = {'http://rinfo.lagrummet.se/ref/rff/nja': 'Högsta domstolen',
            'http://rinfo.lagrummet.se/ref/rff/rh': 'Hovrätterna',
            'http://rinfo.lagrummet.se/ref/rff/rk': 'Kammarrätterna',
            'http://rinfo.lagrummet.se/ref/rff/ra': 'Regeringsrätten',
            'http://rinfo.lagrummet.se/ref/rff/hfd': 'Högsta förvaltningsdomstolen',
            'http://rinfo.lagrummet.se/ref/rff/ad': 'Arbetsdomstolen',
            'http://rinfo.lagrummet.se/ref/rff/fod': 'Försäkringsöverdomstolen',
            'http://rinfo.lagrummet.se/ref/rff/md': 'Marknadsdomstolen',
            'http://rinfo.lagrummet.se/ref/rff/mig': 'Migrationsöverdomstolen',
            'http://rinfo.lagrummet.se/ref/rff/mod': 'Miljööverdomstolen'
            }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

from ferenda import DocumentRepository
from ferenda import util


class PEP(DocumentRepository):
    alias = "pep"
    start_url = "http://hg.python.org/peps"
    document_url_template = "http://hg.python.org/peps/file/tip/pep-%(basefile)s.txt"

    def download(self):
        hg_clone_path = os.sep.join(self.config.datadir, self.alias, 'clone')
        if os.path.exists(hg_clone_path):
            self.log.debug("Pulling latest changes")
            util.runcmd("hg pull", cwd=hg_clone_path)
            self.log.debug("Updating local clone")
            util.runcmd("hg update", cwd=hg_clone_path)
        else:
            hg_clone_parent = os.sep.join(self.config.datadir, self.alias)
            util.runcmd("hg clone %s clone" % self.start_url,
                        cwd=hg_clone_parent)
        new_last_rev = None
        cmd = "LANGUAGE=C hg log -v"
        util.runcmd(cmd)
        for rev in "LANGUAGE=C hg log -v":
            if not new_last_rev:
                new_last_rev = rev.id
            if rev > self.config.last_rev:
                for f in rev.files:  # rev.files only contain proper pep files
                    "hg cat -r %s > downloaded/%s-r%s.txt" % (
                        f, self.store.downloaded_path(f), rev.id)
            else:
                self.config.last_rev = new_last_rev
                break



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# A number of different classes each fetching the same data from
# different sources (and with different data formats and data fidelity)
import os
import re
import functools
import codecs
from datetime import datetime, timedelta
from six.moves.urllib_parse import urljoin

from bs4 import BeautifulSoup
from rdflib import Literal
import requests
from six import text_type as str

from . import SwedishLegalSource, SwedishLegalStore, Trips, Regeringen, RPUBL
from ferenda import Describer
from ferenda import PDFDocumentRepository
from ferenda import CompositeRepository, CompositeStore
from ferenda import TextReader
from ferenda import util
from ferenda.decorators import managedparsing, downloadmax, recordlastdownload
from ferenda.elements import Paragraph
from ferenda.elements import Heading
from ferenda.elements import ListItem
from ferenda.errors import DocumentRemovedError


class Continuation(object):
    pass


class DirTrips(Trips):

    """Downloads Direktiv in plain text format from http://rkrattsbaser.gov.se/dir/"""
    alias = "dirtrips"
    app = "dir"
    base = "DIR"
    # start_url is created (by Trips.download_get_basefiles) from this:
    download_params = [{'maxpage': 101, 'app': app, 'base': base}]

    # overrides Trips.document_url_template
    document_url_template = "http://rkrattsbaser.gov.se/cgi-bin/thw?${APPL}=DIR&${BASE}=DIR&${HTML}=dir_dok&${TRIPSHOW}=format=THW&BET=%(basefile)s"

    rdf_type = RPUBL.Direktiv

    @recordlastdownload
    def download(self, basefile=None):
        if basefile:
            return super(DirTrips, self).download(basefile)
        else:
            if self.config.lastdownload and not self.config.refresh:
                startdate = self.config.lastdownload - timedelta(days=30)
                self.start_url += "&UDAT=%s+till+%s" % (
                    datetime.strftime(startdate, "%Y-%m-%d"),
                    datetime.strftime(datetime.now(), "%Y-%m-%d"))
            super(DirTrips, self).download()

    @managedparsing
    def parse(self, doc):
        # FIXME: need some way of telling intermediate_path that
        # suffix should be .txt (preferably w/o overriding
        # DocumentStore)
        intermediate_path = self.store.path(doc.basefile, 'intermediate', '.txt')
        downloaded_path = self.store.downloaded_path(doc.basefile)
        if not util.outfile_is_newer([downloaded_path], intermediate_path):
            html = codecs.open(downloaded_path, encoding="iso-8859-1").read()
            util.writefile(intermediate_path, util.extract_text(
                html, '<pre>', '</pre>'), encoding="utf-8")
        reader = TextReader(intermediate_path, encoding="utf-8")
        header_chunk = reader.readparagraph()
        self.make_meta(header_chunk, doc.meta, doc.uri, doc.basefile)
        self.make_body(reader, doc.body)

        # Iterate through body tree and find things to link to (See
        # EurlexTreaties.process_body for inspiration)
        self.process_body(doc.body, '', doc.uri)
        return doc

    def header_lines(self, header_chunk):
        n = util.normalize_space
        # This is a ridiculously complicated way of extracting
        # key-value headers when both keys and headers may be
        # continuated. The below, which relies on HTML tags enclosing
        # the value, is much simpler.
        # 
        # header = re.compile("([^:]+):\s*<b>([^<]*)</b>")
        # for m in header.finditer(header_chunk):
        #    yield [util.normalize_space(x) for x in m.groups()]
        ck = cv = ""
        for line in header_chunk.split("\n"):
            if ": " in line:
                # yield buffer
                if ck.strip() and cv.strip():
                    yield(n(ck), n(cv))
                    ck = ""
                k, cv = line.split(":", 1)
                if ck.strip():
                    ck += k
                else:
                    ck = k
            else:
                if line.startswith("    "):
                    cv += line
                else:
                    if ck.strip() and cv.strip():
                        yield(n(ck), n(cv))
                    ck = line
                    cv = ""
        yield(n(ck),n(cv))
                
    def make_meta(self, chunk, meta, uri, basefile):
        d = Describer(meta, uri)
        dct = self.ns['dct']
        prov = self.ns['prov']
        owl = self.ns['owl']
        rpubl = RPUBL

        d.rdftype(self.rdf_type)
        d.value(prov.wasGeneratedBy, self.qualified_class_name())

        # predicates maps key strings to corresponsing RDFLib terms,
        # e.g. "Rubrik" -> dct:title
        predicates = {'Dir nr': dct.identifier,
                      'Departement': rpubl.departement,
                      'Beslut vid regeringssammanträde':
                      rpubl.beslutsdatum,
                      'Rubrik': dct.title,
                      'Senast ändrad': dct.changed
                      }
        # munger contains a set of tuples where the first item is a
        # method for converting a plain text into the appropriate
        # RDFLib value, e.g:
        # - "Utredning av foo" => Literal("Utredning av foo",lang="sv")
        # - "1987-02-19" => datetime(1987,2,19)
        # - "Arbetsdepartementet" => URIRef("http://lagen.nu/terms/arbdep")
        # The second item is the Describer method that
        # should be used to add the value to the graph, i.e. .value
        # for Literals and .rel for URIRefs
        munger = {'Dir nr': (self.sanitize_identifier, d.value),  # the RDFLib constructor
                  'Departement': (functools.partial(self.lookup_resource, warn=False), d.rel),
                  'Beslut vid regeringssammanträde': (self.parse_iso_date, d.value),
                  'Rubrik': (self.sanitize_rubrik, d.value),
                  'Senast ändrad': (self.parse_iso_date, d.value)
                  }
        for (key, val) in self.header_lines(chunk):
            try:
                pred = predicates[key]
                (transformer, setter) = munger[key]
                setter(pred, transformer(val))
            except (KeyError, ValueError):
                self.log.error(
                    "Couldn't munge value '%s' into a proper object for predicate '%s'" % (val, key))

        d.rel(dct.publisher, self.lookup_resource("Regeringskansliet"))
        d.rel(owl.sameAs, self.sameas_uri(uri))
        self.infer_triples(d, basefile)
        # finally, we need a dct:issued, and the best we can come up
        # with is the "Beslut vid regeringssammanträde" date
        # (rpubl:beslutsdatum), so we copy it.
        d.value(dct.issued, d.getvalue(rpubl.beslutsdatum))

    def sanitize_rubrik(self, rubrik):
        if rubrik == "Utgår":
            raise DocumentRemovedError()

        rubrik = re.sub("^/r2/ ", "", rubrik)
        return Literal(rubrik, lang="sv")

    def sanitize_identifier(self, identifier):
        # "Dir.1994:111" -> "Dir. 1994:111"
        if re.match("Dir.\d+", identifier):
            identifier = "Dir. " + identifier[4:]
        if not identifier.startswith("Dir. "):
            identifier = "Dir. " + identifier
        return Literal(identifier)

    def make_body(self, reader, body):
        current_type = None
        for p in reader.getiterator(reader.readparagraph):
            new_type = self.guess_type(p, current_type)
            # if not new_type == None:
            #    print "Guessed %s for %r" % (new_type.__name__,p[:20])
            if new_type is None:
                pass
            elif new_type == Continuation and len(body) > 0:
                # Don't create a new text node, add this text to the last
                # text node created
                para = body.pop()
                para.append(p)
                body.append(para)
            else:
                if new_type == Continuation:
                    new_type = Paragraph
                body.append(new_type([p]))
                current_type = new_type

    def guess_type(self, p, current_type):
        if not p:  # empty string
            return None
        # complex heading detection heuristics: Starts with a capital
        # or a number, and doesn't end with a period (except in some
        # cases).
        elif ((re.match("^\d+", p)
               or p[0].lower() != p[0])
              and not (p.endswith(".") and
                       not (p.endswith("m.m.") or
                            p.endswith("m. m.") or
                            p.endswith("m.fl.") or
                            p.endswith("m. fl.")))):
            return Heading
        elif p.startswith("--"):
            return ListItem
        elif (p[0].upper() != p[0]):
            return Continuation  # magic value, used to glue together
                                 # paragraphs that have been
                                 # inadvertently divided.
        else:
            return Paragraph

    def process_body(self, element, prefix, baseuri):
        if isinstance(element, str):
            return
        fragment = prefix
        uri = baseuri
        for p in element:
            self.process_body(p, fragment, baseuri)

    def canonical_uri(self, basefile):
        return self.config.url + "res/dir/" + basefile

    @classmethod
    def tabs(cls, primary=False):
        return [['Förarbeten', '/forarb/']]


class DirAsp(SwedishLegalSource, PDFDocumentRepository):

    """Downloads Direktiv in PDF format from http://rkrattsdb.gov.se/kompdf/"""
    alias = "dirasp"
    start_url = "http://rkrattsdb.gov.se/kompdf/search.asp"
    document_url = "http://62.95.69.24/KOMdoc/%(yy)02d/%(yy)02d%(num)04d.PDF"
    source_encoding = "iso-8859-1"
    rdf_type = RPUBL.Direktiv

    def download(self):
        resp = requests.get(self.start_url)
        soup = BeautifulSoup(resp.text)
        depts = [opt['value'] for opt in soup.find_all("option", value=True)]
        for basefile, url in self.download_get_basefiles(depts):
            # since the server doesn't support conditional caching and
            # direktivs are basically never updated once published, we
            # avoid even calling download_single if we already have
            # the doc.
            if ((not self.config.refresh) and
                (not os.path.exists(self.store.downloaded_path(basefile)))):
                self.download_single(basefile, url)

    @downloadmax
    def download_get_basefiles(self, depts):
        for dept in depts:
            resp = requests.post(urljoin(self.start_url, 'sql_search_rsp.asp'),
                                 {'departement': dept.encode('latin-1'),
                                  'kom_nr': '',
                                  'title': '',
                                  'ACTION': '  SÖK  '.encode('latin-1')})
            soup = BeautifulSoup(resp.text)
            hits = list(soup.find_all(True, text=re.compile(r'(\d{4}:\d+)')))
            self.log.info("Searching for dept %s, %d results" % (dept, len(hits)))
            for hit in hits:
                link = hit.find_parent("a")
                # convert 2006:02 to 2006:2 for consistency
                segments = re.search("(\d+):(\d+)", link.text).groups()
                basefile = ":".join([str(int(x)) for x in segments])
                # we use link.absolute_url rather than relying on our
                # own basefile -> url code in remote_url. It seems
                # that in least one case the URL formatting rule is
                # not followed by the system...
                yield basefile, urljoin(self.start_url, link['href'])

    def remote_url(self, basefile):
        yy = int(basefile[2:4])
        num = int(basefile[5:])
        return self.document_url % {'yy': yy, 'num': num}

    def canonical_uri(self, basefile):
        return self.config.url + "res/dir/" + basefile

    def parse_from_pdfreader(self, pdfreader, doc):
        super(DirAsp, self).parse_from_pdfreader(pdfreader, doc)
        d = Describer(doc.meta, doc.uri)
        self.infer_triples(d, doc.basefile)
        return doc

    @classmethod
    def tabs(cls, primary=False):
        return [['Förarbeten', '/forarb/']]


class DirPolopoly(Regeringen):

    """Downloads Direktiv in PDF format from http://www.regeringen.se/"""
    alias = "dirpolo"
    cssfiles = ['../ferenda/res/css/pdfview.css']
    jsfiles = ['../ferenda/res/js/pdfviewer.js']
    re_basefile_strict = re.compile(r'Dir\. (\d{4}:\d+)')
    re_basefile_lax = re.compile(r'(?:[Dd]ir\.?|) ?(\d{4}:\d+)')
    rdf_type = RPUBL.Direktiv
    document_type = Regeringen.KOMMITTEDIREKTIV

    def sanitize_identifier(self, identifier):
        # "Dir.1994:111" -> "Dir. 1994:111"
        if re.match("Dir.\d+", identifier):
            identifier = "Dir. " + identifier[4:]
        if not identifier.startswith("Dir. "):
            identifier = "Dir. " + identifier
        return Literal(identifier)

# inherit list_basefiles_for from CompositeStore, basefile_to_pathfrag
# from SwedishLegalStore)
class DirektivStore(CompositeStore, SwedishLegalStore):
    pass

        
# Does parsing, generating etc from base files:
class Direktiv(CompositeRepository, SwedishLegalSource):

    "A composite repository containing ``DirTrips``, ``DirAsp`` and ``DirPolopoly``."""
    subrepos = DirPolopoly, DirAsp, DirTrips
    alias = "dir"
    xslt_template = "paged.xsl"
    xslt_template_toc = "toc.xsl"
    storage_policy = "dir"
    rdf_type = RPUBL.Direktiv
    documentstore_class = DirektivStore
    
    @classmethod
    def tabs(cls, primary=False):
        return [['Förarbeten', '/forarb/']]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# the main idea here is, just like testTriplestore, to just make sure
# every line of code is run once, not to instantiate all different
# implementations/configurations and run them all

import json, re, os
from tempfile import mkstemp, mkdtemp
import shutil

import requests.exceptions

from ferenda import util, errors
from ferenda.compat import patch, Mock, unittest
from ferenda.testutil import FerendaTestCase

# SUT
from ferenda import FulltextIndex
from ferenda import fulltextindex 
from integrationFulltextIndex import WhooshBasicIndex, WhooshBasicQuery
from integrationFulltextIndex import BasicIndex, BasicQuery, ESBase

CREATE_CANNED = False

# this is copied directly from testTriplestore and should perhaps go
# into ferenda.testutil
def canned(*responses, **kwargs):
    returned = []
    param = {}
    def fakeresponse(*args, **kwargs):
        if len(returned) > len(responses):
            raise IndexError("Ran out of canned responses after %s calls" %
                             len(returned))
        resp = Mock()
        resp.status_code = responses[len(returned)][0]
        responsefile = responses[len(returned)][1]
        if responsefile:
            responsefile = "test/files/fulltextindex/" + responsefile
            resp.content = util.readfile(responsefile, "rb")
            resp.text = util.readfile(responsefile)
            if responsefile.endswith(".json"):
                data = json.loads(util.readfile(responsefile))
                resp.json = Mock(return_value=data)
        returned.append(True)
        return resp

    def makeresponse(*args, **kwargs):
        clb = getattr(requests, param['method'])
        resp = clb(*args, **kwargs)
        if resp.status_code != responses[len(returned)][0]:
            print("WARNING: Expected status code %s, got %s (respfile %s)" %
                  (responses[len(returned)][0], resp.status_code,
                   responses[len(returned)][1]))

        responsefile = "test/files/fulltextindex/" + responses[len(returned)][1]
        with open(responsefile, 'wb') as fp:
            fp.write(resp.content)
        returned.append(True)
        return resp

    if kwargs.get('create', True):
        param['method'] = kwargs.get('method')
        return makeresponse
    else:
        return fakeresponse

class MockESBase(ESBase):

    @patch('ferenda.fulltextindex.requests')
    def setUp(self, mock_requests):
        can = canned((404, "exists-not.json"),
                     create=CREATE_CANNED, method="get")
        mock_requests.get.side_effect = can

        can = canned((200, "create.json"),
                     create=CREATE_CANNED, method="post")
        mock_requests.put.side_effect = can
        self.location = "http://localhost:9200/ferenda/"
        self.index = FulltextIndex.connect("ELASTICSEARCH", self.location, [])

    @patch('ferenda.fulltextindex.requests')
    def tearDown(self, mock_requests):
        can = canned((200, "delete.json"),
                     create=CREATE_CANNED, method="delete")
        mock_requests.delete.side_effect = can 
        self.index.destroy()
    
class MockESBasicIndex(BasicIndex, MockESBase):

    @patch('ferenda.fulltextindex.requests')
    def test_create(self, mock_requests):
        # since we stub out MockESBase.setUp (which creates the
        # schema/mapping), the only two requests test_create will do
        # is to check if a mapping exists, and it's definition
        can = canned((200, "exists.json"),
                     (200, "schema.json"),
                     create=CREATE_CANNED, method='get')
        mock_requests.get.side_effect = can
        super(MockESBasicIndex, self).test_create()
        
    @patch('ferenda.fulltextindex.requests')
    def test_insert(self, mock_requests):
        can = canned((201, "insert-1.json"),
                     (201, "insert-2.json"),
                     (201, "insert-3.json"),
                     (200, "insert-4.json"), # no new stuff?
                     create=CREATE_CANNED, method="put")
        mock_requests.put.side_effect = can

        can = canned((200, "commit.json"),
                     (200, "commit.json"),
                     create=CREATE_CANNED, method="post")
        mock_requests.post.side_effect = can

        can = canned((200, "count-2.json"),
                     (200, "count-3.json"),
                     create=CREATE_CANNED, method="get")
        mock_requests.get.side_effect = can

        super(MockESBasicIndex, self).test_insert()

class MockESBasicQuery(BasicQuery, MockESBase): 

    @patch('ferenda.fulltextindex.requests')
    def test_basic(self, mock_requests):
        can = canned((201, "insert-1.json"),
                     (201, "insert-2.json"),
                     (201, "insert-3.json"),
                     (200, "insert-4.json"), # no new stuff?
                     (201, "insert-5.json"),
                     create=CREATE_CANNED, method="put")
        mock_requests.put.side_effect = can

        can = canned((200, "commit.json"),
                     (200, "commit.json"),
                     (200, "commit.json"),
                     (200, "commit.json"),
                     (200, "commit.json"), # one commit per update, because of reasons...
                     (200, "query-main.json"),
                     (200, "query-document.json"),
                     (200, "query-section.json"),
                     create=CREATE_CANNED, method="post")
        mock_requests.post.side_effect = can

        can = canned((200, "count-0.json"),
                     (200, "count-4.json"),
                     create=CREATE_CANNED, method="get")
        mock_requests.get.side_effect = can

        super(MockESBasicQuery, self).test_basic()

    @patch('ferenda.fulltextindex.requests')
    def test_fragmented(self, mock_requests):
        can = canned((201, "insert-1.json"),
                     create=CREATE_CANNED, method="put")
        mock_requests.put.side_effect = can

        can = canned((200, "commit.json"),
                     (200, "query-needle.json"),
                     create=CREATE_CANNED, method="post")
        mock_requests.post.side_effect = can

        super(MockESBasicQuery, self).test_fragmented()

class TestIndexedType(unittest.TestCase):

    def test_eq(self):
        id1 = fulltextindex.Identifier(boost=16)
        id2 = fulltextindex.Identifier(boost=16)
        lbl = fulltextindex.Label(boost=16)
        self.assertEqual(id1, id2)
        self.assertNotEqual(id1, lbl)
    
    def test_repr(self):
        self.assertEqual("<Identifier>", repr(fulltextindex.Identifier()))
        self.assertEqual("<Identifier boost=16>",
                         repr(fulltextindex.Identifier(boost=16)))
        self.assertEqual("<Label boost=16 foo=bar>",
                         repr(fulltextindex.Label(boost=16, foo='bar')))




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys, os
if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())

from ferenda import DocumentRepository, LayeredConfig, util, errors
from ferenda.testutil import RepoTester
#SUT
from ferenda import CompositeRepository

class SubrepoA(DocumentRepository):
    storage_policy = "file"
    alias= "a"
    def download(self, basefile=None):
        util.writefile(self.store.downloaded_path("1"), "basefile 1, repo a")
        util.writefile(self.store.downloaded_path("3"), "basefile 3, repo a")

    def parse(self, basefile):
        if basefile in ("1", "3"):
            util.writefile(self.store.parsed_path(basefile),
                           "basefile %s, parsed by a" % basefile)
            util.writefile(self.store.distilled_path(basefile),
                           "basefile %s, metadata from a" % basefile)
            return True
        else:
            return False # we don't even have this basefile

    def custom(self):
        return False
        
class SubrepoB(DocumentRepository):
    storage_policy = "dir"
    alias= "b"
    def download(self, basefile=None):
        util.writefile(self.store.downloaded_path("1"), "basefile 1, repo b")
        util.writefile(self.store.downloaded_path("2"), "basefile 2, repo b")

    def parse(self, basefile):
        if basefile == "1":
            util.writefile(self.store.parsed_path("1"),
                           "basefile 1, parsed by b")
            util.writefile(self.store.parsed_path("1", attachment="attach.txt"),
                           "attachment for basefile 1, parsed by b")
            util.writefile(self.store.distilled_path("1"),
                           "basefile 1, metadata from b")
            return True
        else:
            raise errors.ParseError("No can do!")

    def get_default_options(self):
        opts = super(SubrepoB, self).get_default_options()
        opts['customproperty'] = "Hello world!"
        return opts

    def custom(self):
        return self.config.customproperty


class CompositeExample(CompositeRepository):
    subrepos = SubrepoB, SubrepoA
    storage_policy = "dir"

    def custom(self):
        for c in self.subrepos:
            inst = self.get_instance(c, self.myoptions)
            ret = inst.custom()
            if ret:
                return ret
        raise RuntimeError("No subrepo could perform custom method")
    

class TestComposite(RepoTester):
    repoclass = CompositeExample
 
    def test_download(self):
        self.repo.download()
        self.assertEqual("basefile 1, repo a",
                         util.readfile(self.datadir+"/a/downloaded/1.html"))
        self.assertEqual("basefile 1, repo b",
                         util.readfile(self.datadir+"/b/downloaded/1/index.html"))
        self.assertEqual("basefile 2, repo b",
                         util.readfile(self.datadir+"/b/downloaded/2/index.html"))
        self.assertEqual("basefile 3, repo a",
                         util.readfile(self.datadir+"/a/downloaded/3.html"))

    def test_list_basefiles_for(self):
        self.repo.download()
        # This doesn't work since self.repo.store.docrepos has
        # uninitialized classes, not objects
        self.assertEqual(set(["3", "2", "1"]),
                         set(self.repo.store.list_basefiles_for("parse")))

    def test_parse(self):
        # we already know list_basefiles_for("parse") will return
        # ["3", "2", "1"]
        self.assertTrue(self.repo.parse("1")) # both A and B can handle this
        # but B should win
        self.assertEqual("basefile 1, parsed by b",
                         util.readfile(self.repo.store.parsed_path("1")))
        self.assertEqual("basefile 1, metadata from b",
                         util.readfile(self.repo.store.distilled_path("1")))
        self.assertEqual(["attach.txt"],
                         list(self.repo.store.list_attachments("1", "parsed")))
        with self.assertRaises(errors.ParseError):
            self.assertFalse(self.repo.parse("2")) # none can handle this
        self.assertTrue(self.repo.parse("3")) # only A can handle this
        self.assertEqual("basefile 3, parsed by a",
                         util.readfile(self.repo.store.parsed_path("3")))
        self.assertEqual("basefile 3, metadata from a",
                         util.readfile(self.repo.store.distilled_path("3")))
        self.assertEqual([], # this repo supports attachment, but
                             # underlying repo A did not
                         list(self.repo.store.list_attachments("3", "parsed")))
                        
        # in this case, all files should be up-to-date, so no copying
        # should occur (triggering the "Attachments are (likely)
        # up-to-date branch")
        self.assertTrue(self.repo.parse("1")) 

        # and finally, list_basefiles_for("generate") should delegate
        # to DocumentStore.list_basefiles_for
        self.assertEqual(set(["1", "3"]),
                         set(self.repo.store.list_basefiles_for("generate")))

    def test_config(self):
        # test it with self.repo being initialized with some kwargs parameters
        self.repo.download()
        got = self.repo.custom()
        self.assertEqual("Hello world!", got)

        # now test with external config object -- this is where it'll fail
#        self.repo = CompositeExample()
#        self.repo.config = LayeredConfig({'datadir': self.datadir})
#        got = self.repo.custom()
#        self.assertEqual("Hello world!", got)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

import sys
import os
import shutil

from ferenda import util, errors

# SUT
from ferenda import PDFDocumentRepository
from ferenda.testutil import RepoTester

if os.getcwd() not in sys.path: sys.path.insert(0,os.getcwd())
from ferenda.manager import setup_logger; setup_logger('CRITICAL')

class Repo(RepoTester):
    repoclass = PDFDocumentRepository
    def test_parse(self):
        util.ensure_dir(self.repo.store.downloaded_path("sample"))
        shutil.copy2("test/files/pdfreader/sample.pdf",
                     self.repo.store.downloaded_path("sample"))
        try:
            self.repo.required_predicates = []
            self.repo.parse("sample")
        except errors.ExternalCommandError:
            # print("pdftohtml error: retrying")
            # for systems that don't have pdftohtml, we copy the expected
            # intermediate files, so that we can test the rest of the logic
            targetdir = os.path.dirname(self.repo.store.intermediate_path("sample"))
            # print("working around by copying test/files/pdfreader/intermediate tree to %s" % targetdir)
            if os.path.exists(targetdir):
                shutil.rmtree(targetdir)
            shutil.copytree("test/files/pdfreader/intermediate",
                            targetdir)
            # make really sure the xml file has a newer timestamp than the PDF
            from time import sleep
            sleep(0.01)
            os.utime(targetdir+"/index.xml", None)
            try:
                self.repo.parse("sample")
            except errors.ExternalCommandError as e:
                print("ExternalCommandError on rerun.\n    targetdir: %s\n    %s exists: %s\n    message: %s" %
                      (targetdir, targetdir+"/index.xml", os.path.exists(targetdir+"/index.xml"), e))
            # print("Workaround succeeded: %s" % os.path.exists(targetdir+"/index.xml"))
        
        p = self.repo.store.datadir
        self.assertTrue(os.path.exists(p+'/intermediate/sample/index001.png'))
        self.assertFalse(os.path.exists(p+'/intermediate/sample/index.pdf'))
        self.assertTrue(os.path.exists(p+'/intermediate/sample/index.xml'))
        self.assertTrue(os.path.exists(p+'/parsed/sample/index001.png'))
        self.assertTrue(os.path.exists(p+'/parsed/sample/index.css'))
        self.assertTrue(os.path.exists(p+'/parsed/sample/index.xhtml'))
    



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import collections
import inspect

import six
from six import text_type as str

from ferenda.errors import FSMStateError


class FSMParser():

    """A configurable finite state machine (FSM) for parsing documents
    with nested structure. You provide a set of *recognizers*, a set
    of *constructors*, a *transition table* and a *stream* of document
    text chunks, and it returns a hierarchical document object
    structure.

    See :doc:`../fsmparser`.

    """

    def __init__(self):
        self.debug = False
        self.transitions = None  # set by set_transitions
        self.recognizers = None  # set by set_recognizers() or set_transitions()
        self.reader = None  # set by parse()
        # somewhat magic
        self.initial_state = None
        self.initial_constructor = None
        # pseudo-internal
        self._state_stack = []

    def _debug(self, msg):
        """Prints a debug message, indented to show how far down in the nested structure we are"""
        if self.debug:
            stack = inspect.stack()
            calling_frame = [x[3] for x in stack][1]
            relative_depth = len(self._state_stack)
            # print("%s[%s(%r)] %s" % (". " * relative_depth, calling_frame, self._state_stack, msg))
            state = "/".join(self._state_stack)
            print("%s/%s(): %s" % (state, calling_frame, msg))

    def set_recognizers(self, *args):
        """Set the list of functions (or other callables) used in
        order to recognize symbols from the stream of text
        chunks. Recognizers are tried in the order specified here."""
        self.recognizers = args

    def set_transitions(self, transitions):
        """Set the transition table for the state matchine.

        :param transitions: The transition table, in the form of a mapping between two tuples. The first tuple should be the current state (or a list of possible current states) and a callable function that determines if a particular symbol is recognized ``(currentstate, recognizer)``. The second tuple should be a constructor function (or `False```) and the new state to transition into.
        
        """
        self.transitions = {}
        for (before, after) in transitions.items():
            (before_states, recognizer) = before
            if not callable(after):
                (constructor, after_state) = after
                assert (constructor == False) or callable(
                    constructor), "Specified constructor %r not callable" % constructor
            assert callable(recognizer), "Specified recognizer %r not callable" % recognizer
            if (not isinstance(before_states, (list, tuple))):
                before_states = [before_states]
            for before_state in before_states:
                if callable(after):
                    self._debug("%r,%s() -> %s()" %
                                (before_state, recognizer.__name__, after.__name__))
                elif callable(after[0]):
                    self._debug("%r,%s() -> %s(), %r" %
                                (before_state, recognizer.__name__, after[0].__name__, after[1]))
                else:
                    self._debug("%r,%s() -> %r, %r" %
                                (before_state, recognizer.__name__, after[0], after[1]))
                self.transitions[(before_state, recognizer)] = after

    def parse(self, chunks):
        """Parse a document in the form of an iterable of suitable
        chunks -- often lines or elements.  each chunk should be a
        string or a string-like obje ct.  Some examples::
        
          p = FSMParser()
          reader = TextReader("foo.txt")
          body = p.parse(reader.getiterator(reader.readparagraph),"body", make_body)
          body = p.parse(BeautifulSoup("foo.html").find_all("#main p"), "body", make_body)
          body = p.parse(ElementTree.parse("foo.xml").find(".//paragraph"), "body", make_body)

        :param chunks: The document to be parsed, as a list or any other
                       iterable of text-like objects.
        :param initialstate: The initial state for the machine. The
                             state must be present in the transition
                             table. This could be any object, but strings are
                             preferrable as they make error messages
                             easier to understand.
        :param initialconstructor: A function that creates a document
                                   root object, and then fills it with
                                   child objects using
                                   .make_children()
        :type initialconstructor: callable
        :returns: A document object tree.
        """
        self._debug("Starting parse")
        self.reader = Peekable(chunks)
        self._state_stack = [self.initial_state]
        return self.initial_constructor(self)

    def analyze_symbol(self):
        """Internal function used by make_children()"""
        try:
            rawchunk = self.reader.peek()
            # chunk = str(rawchunk)
            chunk = repr(rawchunk)
            if len(chunk) > 90:
                # chunk = chunk[:25] + "[...]" + chunk[-10:]
                seg = (chunk[:25], chunk[-10:])
                try:
                    chunk = "%s [...] %s" % seg
                except UnicodeDecodeError:
                    chunk = "%r [...] %r" % seg
            else:
                chunk = chunk

        except StopIteration:
            self._debug("We're done!")
            return None

        applicable_tmp = [x[1] for x in self.transitions.keys() if x[0] == self._state_stack[-1]]
        # Create correct sorting of applicable_recognizers
        applicable_recognizers = []
        for recognizer in self.recognizers:
            if recognizer in applicable_tmp:
                applicable_recognizers.append(recognizer)

        applicable_display = ", ".join([x.__name__ for x in applicable_recognizers])
        for recognizer in applicable_recognizers:
            if recognizer(self):
                self._debug("Tested %r against %s -> %s " %
                            (chunk, applicable_display,
                             recognizer.__name__))
                # self._debug("%r -> %s" % (chunk, recognizer.__name__))
                return recognizer
        raise FSMStateError("No recognizer match for %s (tried %s)" % (chunk, applicable_display))

    def transition(self, currentstate, symbol):
        """Internal function used by make_children()"""
        assert (currentstate, symbol) in self.transitions, "(%r, %r) should be in self.transitions" % (currentstate, symbol)

        t = self.transitions[(currentstate, symbol)]
        if callable(t):
            return t(symbol, self._state_stack)
        else:
            return t

    def make_child(self, constructor, childstate):
        """Internal function used by make_children(), which calls one
        of the constructors defined in the transition table."""

        if not childstate:
            childstate = self._state_stack[-1]
            # self._debug("calling child constructor %s" % constructor.__name__)
        else:
            # self._debug("calling child constructor %s in state %r" %
            #             (constructor.__name__, childstate))
            pass
        self._state_stack.append(childstate)
        ret = constructor(self)
        self._state_stack.pop()  # do something with this?
        return ret

    def make_children(self, parent):
        """Creates child nodes for the current (parent) document node.

        :param parent: The parent document node, as any list-like object
                       (preferrably a subclass of
                       :py:class:`ferenda.elements.CompoundElement`)
        :returns: The same ``parent`` object.
        
        """
        self._debug("Making children for %s" % parent.__class__.__name__)
        while True:  # we'll break out of this when transition()
                    # returns a constructor that is False
            symbol = self.analyze_symbol()
            if symbol is None:  # no more symbols
                self._debug("We're done!")
                return parent

            (constructor, newstate) = self.transition(self._state_stack[-1],
                                                      symbol)

            if constructor is False:
                self._debug("transition(%r,%s()) -> (False,%r)" %
                            (self._state_stack[-1], symbol.__name__, newstate))
            else:
                self._debug("transition(%r,%s()) -> (%s(),%r)" %
                            (self._state_stack[-1], symbol.__name__,
                             constructor.__name__, newstate))

            # if transition() indicated that we should change state,
            # first find out whether the constructor will call
            # make_child, creating a new stack frame. This is
            # indicated by the callable having the 'newstate'
            # attribute (set by the @ferenda.decorators.newstate
            # decorator)
            if newstate and not hasattr(constructor, 'newstate'):
                self._debug("Changing top of state stack (%r->%r)" %
                            (self._state_stack[-1], newstate))
                self._state_stack[-1] = newstate

            if constructor:
                element = self.make_child(constructor, newstate)
                if element is not None:
                    parent.append(element)
            else:
                # special weird hack - set the state we'll be
                # returning to by manipulating self._state_stack
                # FIXME: we have no regular test case for this path,
                # but integrationRFC excercises it
                if newstate:
                    self._debug("Changing the state we'll return to (self._state_stack[-2])")
                    self._debug("  (from %r to %r)" % (self._state_stack[-2], newstate))
                    self._state_stack[-2] = newstate
                return parent


# inspired by recipe 19.18 in the python cookbook. A implementation detail helper for FSMParser.
class Peekable(six.Iterator):

    def __init__(self, iterable):
        self._iterable = iter(iterable)
        self._cache = collections.deque()

    def __iter__(self):
        return self

    def _fillcache(self):
        while len(self._cache) < 1:
            self._cache.append(six.advance_iterator(self._iterable))

    def __next__(self):
        self._fillcache()
        result = self._cache.popleft()
        return result

    # useful alias
    next = __next__

    def peek(self):
        self._fillcache()
        result = self._cache[0]
        return result



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

from ferenda import DocumentRepository, DocumentStore
from ferenda import util, errors

class CompositeStore(DocumentStore):

    def __init__(self, datadir, downloaded_suffix=".html",
                 storage_policy="file",
                 docrepo_instances=None):
        self.datadir = datadir  # docrepo.datadir + docrepo.alias
        self.downloaded_suffix = downloaded_suffix
        self.storage_policy = storage_policy
        if not docrepo_instances:
            docrepo_instances = {}
        self.docrepo_instances = docrepo_instances

    def list_basefiles_for(self, action, basedir=None):
        if not basedir:
            basedir = self.datadir
        if action == "parse":
            documents = set()
            for cls, inst in self.docrepo_instances.items():
                for basefile in inst.store.list_basefiles_for("parse"):
                    if basefile not in documents:
                        documents.add(basefile)
                        yield basefile
        else:
            for basefile in super(CompositeStore, self).list_basefiles_for(action):
                yield basefile


class CompositeRepository(DocumentRepository):
    subrepos = ()  # list of classes
    documentstore_class = CompositeStore

    def get_instance(self, instanceclass, options={}):
        if not instanceclass in self._instances:
            if options:
                inst = instanceclass(**options)
            else:
                inst = instanceclass()
                inst.config = self.config 
            self._instances[instanceclass] = inst
        return self._instances[instanceclass]

    def __init__(self, **kwargs):
        self._instances = {}
        self.myoptions = kwargs
        super(CompositeRepository, self).__init__(**kwargs)
        # At this point, self._instances is empty. And we can't really
        # populate it at this time, because we need access to the
        # config object that manager._run_class will set after
        # __init__ finishes, in order to properly initialize all our
        # subrepos.

        # when using API, we won't need an externally-provided config
        # object, as whatever is passed in as **kwargs will populate
        # an internallly-constructed config object. So if that's the
        # case, let's go ahead and create instances and make a store
        # right now.
        if self.myoptions:
            for c in self.subrepos: # populate self._instances
                self.get_instance(c, self.myoptions)
                
            cls = self.documentstore_class
            self.store = cls(self.config.datadir + os.sep + self.alias,
                             downloaded_suffix=self.downloaded_suffix,
                             storage_policy=self.storage_policy,
                             docrepo_instances=self._instances)
        else:
            # no **kwargs were provided, delay the creation of
            # self._instances (and self.store) until we have a proper
            # config object (in the config.setter decorated stuff
            pass
            

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config):
        self._config = config

        for c in self.subrepos: # populate self._instances
            self.get_instance(c)

        cls = self.documentstore_class
        self.store = cls(self.config.datadir + os.sep + self.alias,
                         downloaded_suffix=self.downloaded_suffix,
                         storage_policy=self.storage_policy,
                         docrepo_instances=self._instances)
        

    def download(self):
        for c in self.subrepos:
            inst = self.get_instance(c, self.myoptions)
            # make sure that our store has access to our now
            # initialized subrepo objects
            if c not in self.store.docrepo_instances:
                self.store.docrepo_instances[c] = inst
            inst.download()

    # NOTE: this impl should NOT use the @managedparsing decorator
    def parse(self, basefile):
        # first, check if we really need to parse. If any subrepo
        # returns that parseneeded is false and we have parsed file in
        # the mainrepo, then we're done. This is mainly to avoid the
        # log message below (to be in line with expected repo
        # behaviour of not logging anythin at severity INFO if no real
        # work was done), it does not noticably affect performance
        force = (self.config.force is True or
                 self.config.parseforce is True)
        if not force:
            for c in self.subrepos:
                inst = self.get_instance(c, self.myoptions)
                needed = inst.parseneeded(basefile)
                if not needed and os.path.exists(self.store.parsed_path(basefile)):
                    self.log.debug("%s: Skipped" % basefile)
                    return True # signals everything OK

        with util.logtime(self.log.info, "%(basefile)s OK (%(elapsed).3f sec)",
                          {'basefile': basefile}):
            ret = False
            for c in self.subrepos:
                inst = self.get_instance(c, self.myoptions)
                try:
                    # each parse method should be smart about whether
                    # to re-parse or not (i.e. use the @managedparsing
                    # decorator).
                    ret = inst.parse(basefile)

                except Exception as e: # Any error thrown (errors.ParseError or something else) means we try next subrepo
                    self.log.debug("%s: parse with %s failed: %s" % (basefile, inst.qualified_class_name(), str(e)))
                    ret = False
                if ret:
                    break
            if ret:
                self.copy_parsed(basefile, inst)
        if ret:
            return ret
        else:
            raise errors.ParseError("No instance of %r was able to parse %s" % (self.subrepos, basefile))


    def copy_parsed(self, basefile, instance):
        # If the distilled and parsed links are recent, assume that
        # all external resources are OK as well
        if (util.outfile_is_newer([instance.store.distilled_path(basefile)],
                                  self.store.distilled_path(basefile)) and
            util.outfile_is_newer([instance.store.parsed_path(basefile)],
                                  self.store.parsed_path(basefile))):
            self.log.debug("%s: Attachments are (likely) up-to-date" % basefile)
            return

        util.link_or_copy(instance.store.distilled_path(basefile),
                          self.store.distilled_path(basefile))

        util.link_or_copy(instance.store.parsed_path(basefile),
                          self.store.parsed_path(basefile))

        cnt = 0
        if instance.store.storage_policy == "dir":
            for attachment in instance.store.list_attachments(basefile, "parsed"):
                cnt += 1
                src = instance.store.parsed_path(basefile, attachment=attachment)
                target = self.store.parsed_path(basefile, attachment=attachment)
                util.link_or_copy(src, target)
            if cnt:
                self.log.debug("%s: Linked %s attachments from %s to %s" %
                               (basefile,
                                cnt,
                                os.path.dirname(instance.store.parsed_path(basefile)),
                                os.path.dirname(self.store.parsed_path(basefile))))



# flake8: noqa
from rdflib import Namespace
RPUBL = Namespace('http://rinfo.lagrummet.se/ns/2008/11/rinfo/publ#')
from .swedishlegalsource import SwedishLegalSource, SwedishLegalStore, SwedishCitationParser
from .regeringen import Regeringen
from .riksdagen import Riksdagen
from .trips import Trips, NoMoreLinks
from .arn import ARN
from .direktiv import Direktiv
from .ds import Ds
from .dv import DV
from .jk import JK
from .jo import JO
from .kommitte import Kommitte
from .myndfskr import MyndFskr
from .propositioner import Propositioner
from .sfs import SFS
from .sou import SOU



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re


from . import Regeringen, RPUBL

# See SOU.py for discussion about possible other sources


class Ds(Regeringen):
    alias = "ds"
    re_basefile_strict = re.compile(r'Ds (\d{4}:\d+)')
    re_basefile_lax = re.compile(r'(?:Ds|) ?(\d{4}:\d+)', re.IGNORECASE)
    rdf_type = RPUBL.Utredning
    document_type = Regeringen.DS



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


class TocCriteria(object):

    """Represents a particular way of organizing the documents in a
    repository, for the purpose of generating a table of contents for
    those douments.

    :param binding: The variable name (binding) for that same
                    predicate in the sparql results retrieved from the
                    query constructed by
                    :py:meth:`~ferenda.DocumentRepository.toc_query`.
    :type  binding: str
    :param label: A text label for the entire set of toc for this criteria.
    :type  label: str
    :param pagetitle: A template string used together with whatever
                      selector returns to form a full TOC page title,
                      eg ``Documents starting with "%s"``.
    :type  pagetitle: str
    :param selector: Function that takes a single dict (from
                     :py:meth:`~ferenda.DocumentRepository.toc_select`)
                     and returns a single value used for grouping
                     documents (eg. first letter of ``title``,
                     year-part of ``date``, etc).
    :type  selector: callable
    :param key: Function that takes a single dict and returns a value
                that can be used for sorting documents
    :type key: callable
    :param selector_descending: Whether pagesets constructed by this selector should be sorted in descending (reverse) order
    :type  selector_descending: bool
    :param key_descending: Whether pages selected by this key should be sorted in descending (reverse) order
    :type  key_descending: bool
    """

    def __init__(self, binding, label, pagetitle, selector, key, selector_descending=False, key_descending=False):
        self.binding = binding
        self.label = label
        self.pagetitle = pagetitle
        self.selector = selector
        self.key = key
        self.selector_descending = selector_descending
        self.key_descending = key_descending



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# Base class for fetching data from an ancient database system used by
# swedish gov IT...
import re
from six.moves.urllib_parse import quote

import requests
import lxml.html
from bs4 import BeautifulSoup

from ferenda.decorators import downloadmax, recordlastdownload
from ferenda import util
from . import SwedishLegalSource


class NoMoreLinks(Exception):

    def __init__(self, nextpage=None):
        super(NoMoreLinks, self).__init__()
        self.nextpage = nextpage


class Trips(SwedishLegalSource):
    alias = None  # abstract class
    basefile_regex = "(?P<basefile>\d{4}:\d+)$"

    app = None  # komm, dir, prop, sfst
    base = None  # KOMM, DIR, THWALLPROP, SFSR

    source_encoding = "iso-8859-1"

    # NOTE: both SFS and direktiv.DirTrips override this -- hard to find a
    # template that works for everyone
    document_url_template = ("http://rkrattsbaser.gov.se/cgi-bin/thw?"
                             "${HTML}=%(app)s_lst"
                             "&${OOHTML}=%(app)s_doc&${BASE}=%(base)s"
                             "&${TRIPSHOW}=format=THW&BET=%(basefile)s")
    start_url = ("http://rkrattsbaser.gov.se/cgi-bin/thw?${HTML}=%(app)s_lst"
                 "&${OOHTML}=%(app)s_doc&${SNHTML}=%(app)s_err"
                 "&${MAXPAGE}=%(maxpage)d&${TRIPSHOW}=format=THW"
                 "&${BASE}=%(base)s")

    # for SFS
    # start_url = ("http://rkrattsbaser.gov.se/cgi-bin/thw?${HTML}=%(app)s_lst"
    #              "&${OOHTML}=%(app)s_dok&${SNHTML}=%(app)s_err"
    #              "&${MAXPAGE}=%(maxpage)d&${BASE}=%(base)s"
    #             "&${FORD}=FIND&%%C5R=FR%%C5N+%(start)s&%%C5R=TILL+%(end)s")

    download_params = [{'maxpage': 101, 'app': app, 'base': base}]
    # for SFS:
    # download_params = [{'maxpage': 101,
    #                     'app': app,
    #                     'base': base,
    #                     'start': '1600',
    #                     'end': '2008'},
    #                    {'maxpage': 101,
    #                     'app': app,
    #                     'base': base,
    #                     'start': '2009',
    #                     'end': str(datetime.today().year)}]
    def download(self, basefile=None):
        if basefile:
            return self.download_single(basefile)
        for basefile, url in self.download_get_basefiles(self.download_params):
            self.download_single(basefile, url)

    @downloadmax
    def download_get_basefiles(self, params):
        for param in params:
            done = False
            url = self.start_url % param
            pagecount = 1
            while not done:
                self.log.info("Starting at %s" % url)
                resp = requests.get(url)
                tree = lxml.html.document_fromstring(resp.text)
                tree.make_links_absolute(url, resolve_base_href=True)
                try:
                    for basefile, url in self.download_get_basefiles_page(tree):
                        yield basefile, url
                except NoMoreLinks as e:
                    if e.nextpage:
                        pagecount += 1
                        url = e.nextpage
                        self.log.info("Getting page #%s of results" % pagecount)
                    else:
                        done = True

    def download_get_basefiles_page(self, pagetree):
        nextpage = None
        for element, attribute, link, pos in pagetree.iterlinks():
            if element.text is None:
                continue
            m = re.search(self.basefile_regex, element.text)
            if m:
                basefile = m.group("basefile")
                docurl = link
            else:
                basefile = docurl = None

            if basefile:
                yield(basefile, docurl)
            else:
                # maybe this is the "next page" link?
                m = re.match("Fler poster", element.text)
                if m:
                    nextpage = link
        raise NoMoreLinks(nextpage)

    def download_single(self, basefile, url=None):
        # explicitly call superclass' download_single WITHOUT url
        # parameter. The reason is so that we construct the url
        # through self.remote_url, which provides permanent urls to
        # the wanted documents, instead of the temporary/session id
        # based urls that download_get_basefiles can provide
        return super(Trips, self).download_single(basefile)

    def download_is_different(self, existing, new):
        # load both existing and new into a BeautifulSoup object, then
        # compare the first <pre> element
        existing_soup = BeautifulSoup(util.readfile(existing, encoding=self.source_encoding))
        new_soup = BeautifulSoup(util.readfile(new, encoding=self.source_encoding))
        return existing_soup.pre != new_soup.pre

    def remote_url(self, basefile):
        return self.document_url_template % {'basefile': quote(basefile),
                                             'app': self.app,
                                             'base': self.base}
    def canonical_uri(self, basefile):
        seg = {'dir': 'dir',
               'prop': 'prop',
               'sfst': 'sfs'}
        return self.config.url + "res/%s/%s" % (seg[self.app], basefile)
